from datetime import datetime, timedelta

from pyaerocom.aeroval import EvalSetup
from pyaerocom.scripts.cams2_83.processer import CAMS2_83_Processer
from tests.conftest import lustre_unavail


@lustre_unavail
def test__run_single_entry(patched_config, tmp_path):
    setup = EvalSetup(**patched_config)
    processer = CAMS2_83_Processer(setup)
    processer.cfg._check_time_config()
    assert processer.cfg.colocation_opts.start == "2025/03/01 00:00:00"

    col = processer.get_colocator("EMEP", "EEA")
    startstring = datetime.strptime(
        col.colocation_setup.model_kwargs["daterange"][0], "%Y-%m-%d"
    ).strftime("%Y%m%d")
    stopstring = datetime.strptime(
        col.colocation_setup.model_kwargs["daterange"][1], "%Y-%m-%d"
    ).strftime("%Y%m%d")
    newstartstring = (
        datetime.strptime(col.colocation_setup.model_kwargs["daterange"][0], "%Y-%m-%d")
        - timedelta(days=1)
    ).strftime("%Y%m%d")

    var = "concno2"
    model_name = "EMEP"
    processer._run_single_entry(
        model_name=model_name, obs_name="EEA", var_list=[var], analysis=False
    )

    assert col.colocation_setup.model_name == model_name
    d0 = (
        tmp_path
        / f"{processer.cfg.proj_id}/{processer.cfg.exp_id}/CAMS2-83-{model_name}-day0-FC/{var}_{var}_MOD-CAMS2-83-{model_name}-day0-FC_REF-EEA-UTD_{startstring}_{stopstring}_hourly_ALL-wMOUNTAINS.nc"
    )
    dp = (
        tmp_path
        / f"{processer.cfg.proj_id}/{processer.cfg.exp_id}/CAMS2-83-{model_name}-persistence-FC/{var}_{var}_MOD-CAMS2-83-{model_name}-persistence-FC_REF-EEA-UTD_{newstartstring}_{stopstring}_hourly_ALL-wMOUNTAINS.nc"
    )

    assert d0.is_file()
    assert dp.is_file()



from pydantic import BaseModel, ValidationError, model_validator

from typing import Literal

from pyaerocom import const


class ClimatologyConfig(BaseModel):
    """
    Holds the configuration for the climatology

    Attributes
    -------------
    start : int, optional
        Start year of the climatology
    stop : int, optional
        Stop year of the climatology
    resample_how : str, optional
        How to resample the climatology. Must be mean or median.
    freq : str, optional
        Which frequency the climatology should have
    min_count : dict, optional
        Number of values should be present for the data to be used in the climatology.
        Dict where freqs are the keys and the count is the values

    """

    start: int = const.CLIM_START
    stop: int = const.CLIM_STOP

    set_year: int | None = None

    resample_how: Literal["mean", "median"] = const.CLIM_RESAMPLE_HOW
    freq: str = const.CLIM_FREQ
    min_count: dict = const.CLIM_MIN_COUNT

    @model_validator(mode="after")
    def validate_set_year(self):
        if self.set_year is None:
            self.set_year = int((self.stop - self.start) // 2 + self.start) + 1

        if self.set_year > 2100:
            raise ValidationError
        return self



#!/usr/bin/env python3

import matplotlib.pyplot as plt

import pyaerocom as pya

plt.close("all")

from pathlib import Path

import pyaerocom.testdata_access as td
from pyaerocom.conftest import CHECK_PATHS

tda = td.TestDataAccess()

TESTDATADIR = Path(tda.testdatadir)

OUTBASE = TESTDATADIR.joinpath("coldata")

if not OUTBASE.exists():
    OUTBASE.mkdir()

fpath = TESTDATADIR.joinpath(CHECK_PATHS["tm5aod"])
if not fpath.exists():
    raise Exception("Unexpected error, please debug")
mod = pya.GriddedData(fpath)

obs = pya.io.ReadAeronetSunV3("AeronetSunV3L2Subset.daily").read("od550aer")

coldata = pya.colocation.colocate_gridded_ungridded(mod, obs)

coldata.to_netcdf(OUTBASE)

print(coldata.calc_statistics())

coldata.plot_coordinates()

mod = mod.sel(latitude=(0, 3), longitude=(0, 4))
cgg = pya.colocation.colocate_gridded_gridded(mod, mod)
cgg.data = cgg.data[:, :3]

cgg.plot_scatter()

cgg.to_netcdf(OUTBASE)

pya.plot.mapping.plot_nmb_map_colocateddata(cgg)



import glob
import logging
import os
from pathlib import Path

import numpy as np
import pandas as pd
import xarray as xr

from pyaerocom.exceptions import DataSourceError
from pyaerocom.helpers import varlist_aerocom
from pyaerocom.io.readungriddedbase import ReadUngriddedBase
from pyaerocom.metastandards import StationMetaData
from pyaerocom.units.datetime import TsType
from pyaerocom.ungriddeddata import UngriddedData

from .additional_variables import vmr_to_ghost_stations
from .meta_keys import ghost_meta_keys

from pyaerocom.units import convert_unit

logger = logging.getLogger(__name__)


class ReadGhost(ReadUngriddedBase):
    """Reading interface for GHOST data

    This reader supports both reading of provided EEA and EBAS datasets.
    It may not have implemented all variables provided in the input data, you
    can check all provided variables via :attr:`PROVIDES_VARIABLES`. If you
    find one variable missing in this reading routine, you can register it
    yourself in the class header (see :attr:`VARNAMES_DATA` below) from a
    forked pyaerocom repo and send a pull request, or you can contact us about
    it.

    Note
    ----
    This class inherits from the metaclass template `ReadUngriddedBase`.
    Please have a look at that and make sure you understand the idea behind
    it.

    Information about data quality (Email from D. Bowdalo 2.3.2020)
    --------------------------------------------------------------
    Inside is hourly and daily resolution data for O3, NO, NO2, CO, SO2, PM2.5
    and PM10 for 2018 and 2019.

    The data is a mesh of the EEA E1a (validated)/E2a (UTD) data streams.
    Almost all of the 2018 data is from the E1a stream and all of the 2019 is
    from the E2a stream.
    These can be entirely separated out using the QA code 4 ('Not Maximum
    Data Quality Level’), for which all data from the E2a stream is flagged
    with.

    Additionally I put the QA code 5 ('Preliminary Data’) in my list of
    default QA codes to screen by Jonas, but you may wish to remove this as a
    lot of the 2019 E2a data is flagged by EEA as preliminary, and therefore
    flagged by my processing accordingly.
    """

    #: version of the reader
    __version__ = "0.0.13"

    _FILEMASK = "*.nc"

    #: Default data ID
    DATA_ID = "GHOST.EEA.daily"

    #: IDs of supported datasets
    SUPPORTED_DATASETS = [
        "GHOST.EEA.monthly",
        "GHOST.EEA.hourly",
        "GHOST.EEA.daily",
        "GHOST.EBAS.monthly",
        "GHOST.EBAS.hourly",
        "GHOST.EBAS.daily",
    ]

    #: sampling frequencies of supported datasets
    TS_TYPES = {
        "GHOST.EEA.monthly": "monthly",
        "GHOST.EEA.hourly": "hourly",
        "GHOST.EEA.daily": "daily",
        "GHOST.EBAS.monthly": "monthly",
        "GHOST.EBAS.hourly": "hourly",
        "GHOST.EBAS.daily": "daily",
    }

    #: List of GHOST metadata keys
    META_KEYS = ghost_meta_keys()

    #: Names of flag variables in GHOST NetCDF files
    FLAG_VARS = ["flag", "qa"]

    #:
    FLAG_DIMNAMES = {"qa": "N_qa_codes", "flag": "N_flag_codes"}

    #: dictionary mapping GHOST variable names to AeroCom variable names
    VARNAMES_DATA = {
        "concpm10": "pm10",
        "concpm10al": "pm10al",
        "concpm10as": "pm10as",
        "concpm25": "pm2p5",
        "concpm1": "pm1",
        "conccl": "sconccl",
        "concso4": "sconcso4",
        "vmrco": "sconcco",
        "vmrno": "sconcno",
        "vmrno2": "sconcno2",
        "vmro3": "sconco3",
        "vmrso2": "sconcso2",
        "concno3": "sconcno3",
        "concnh4": "sconcnh4",
    }

    AUX_REQUIRES = {
        "concco": ["vmrco"],
        "concno": ["vmrno"],
        "concno2": ["vmrno2"],
        "conco3": ["vmro3"],
        "concso2": ["vmrso2"],
    }

    AUX_FUNS = {
        "concco": vmr_to_ghost_stations,
        "concno": vmr_to_ghost_stations,
        "concno2": vmr_to_ghost_stations,
        "conco3": vmr_to_ghost_stations,
        "concso2": vmr_to_ghost_stations,
    }

    CONVERT_UNITS_META = {
        "network_provided_volume_standard_pressure": "Pa",
    }

    # This is the default list of flags that mark bad / invalid data, as
    # provided by Dene: [0, 1, 2, 3, 6, 20, 21, 22, 72, 75, 82, 83, 90, 91,
    # 92, 105 (removed), 110, 111, 112, 113, 115, 132, 133]

    #: Default flags used to invalidate data points (these may be either from
    #: provided flag or qa variable, or both, currently only from qa variable)
    DEFAULT_FLAGS_INVALID = {
        "qa": np.asarray(
            [
                [
                    0,
                    1,
                    2,
                    3,
                    6,
                    20,
                    21,
                    22,
                    72,
                    75,
                    82,
                    83,
                    90,
                    91,
                    92,
                    110,
                    111,
                    112,
                    113,
                    115,
                    131,
                    132,
                    133,
                ]
            ]
        ),
        "flag": None,
    }

    @property
    def PROVIDES_VARIABLES(self):
        """
        list of variable names that can be retrieved through this interface
        """
        return list(self.VARNAMES_DATA) + list(self.AUX_FUNS)

    @property
    def DEFAULT_VARS(self):
        """
        list of default variables to retrieve
        """
        return self.PROVIDES_VARIABLES

    @property
    def var_names_data_inv(self):
        """
        Inverted version of dictionary :attr:`VARNAMES_DATA`
        """
        try:
            return self._var_names_inv
        except AttributeError:
            self._var_names_inv = {v: k for k, v in self.VARNAMES_DATA.items()}
            return self._var_names_inv

    @property
    def TS_TYPE(self):
        """
        Default implementation of string for temporal resolution
        """
        try:
            return self.TS_TYPES[self.data_id]
        except KeyError:
            return self._ts_type_from_data_dir()

    def get_file_list(self, vars_to_read=None, pattern=None):
        """
        Retrieve a list of files to read based on input variable names

        Parameters
        ----------
        vars_to_read : str, optional
            list of variables to be imported. If None, use The default is None.
        pattern : TYPE, optional
            DESCRIPTION. The default is None.

        Raises
        ------
        ValueError
            If no files can be found for any of the input variables.

        Returns
        -------
        list
            list with file paths

        """
        if vars_to_read is None:
            vars_to_read = self.PROVIDES_VARIABLES
        elif isinstance(vars_to_read, str):
            vars_to_read = [vars_to_read]

        if pattern is None:
            pattern = self._FILEMASK

        files = []
        for var in vars_to_read:
            if var in self.VARNAMES_DATA:
                # make sure to check for right variable, user may use either
                # AeroCom variable name or GHOST variable name
                var = self.VARNAMES_DATA[var]

            _dir = os.path.join(self.data_dir, var)
            _files = glob.glob(f"{_dir}/{pattern}")
            if not _files:  # pragma: no cover
                raise DataSourceError(f"Could not find any data files for {var}")
            files.extend(_files)

        self.files = sorted(files)
        return self.files

    def get_meta_filename(self, filename):
        """Extract metadata from data filename

        Parameters
        ----------
        filename : str
            data file path or name.

        Returns
        -------
        dict
            dictionary containing var_name, start and stop, and eventually
            also frequency (ts_type)
        """
        var, time = os.path.basename(filename).split(".nc")[0].split("_")

        per = pd.Period(freq="M", year=int(time[:4]), month=int(time[-2:]))
        return dict(var_name=var, start=per.start_time, stop=per.end_time)

    @staticmethod
    def _eval_flags_slice(slc, invalid_flags):
        """
        Compare a flag slice of a data point with input flags marking invalid

        Returns
        -------
        bool
            True, if data point is valid, else False
        """
        if len(np.intersect1d(slc, invalid_flags)) == 0:
            return True
        return False

    def _ts_type_from_data_dir(self):
        try:
            freq = str(TsType(os.path.basename(self.data_dir)))
        except Exception:  # pragma: no cover
            freq = "undefined"
        self.TS_TYPES[self.data_id] = freq
        return freq

    def _eval_flags(self, vardata, invalidate_flags, ds):
        valid = np.ones_like(vardata).astype(bool)
        for flagvar in self.FLAG_VARS:
            # check if this flag variable is in input dictionary
            if flagvar in invalidate_flags:
                invalidate = invalidate_flags[flagvar]
                if invalidate is None:
                    continue
                flags = ds[flagvar]
                slice_dim = flags.dims.index(self.FLAG_DIMNAMES[flagvar])

                valid *= np.apply_along_axis(
                    self._eval_flags_slice, slice_dim, flags.values, invalidate
                )
        invalid = ~valid
        return invalid

    def read_file(self, filename, var_to_read=None, invalidate_flags=None, var_to_write=None):
        """Read GHOST NetCDF data file

        Parameters
        ----------
        filename : str
            absolute path to filename to read
        var_name : str, optional
            name of variable to be read, if None, it is inferred from filename

        Returns
        -------
        list
            list of loaded `StationData` objects (dict-like data objects)

        """
        if invalidate_flags is None:
            invalidate_flags = self.DEFAULT_FLAGS_INVALID

        if var_to_read is None:
            var_to_read = self.get_meta_filename(filename)["var_name"]
        elif var_to_read in self.VARNAMES_DATA:
            if var_to_write is None:
                var_to_read, var_to_write = self.VARNAMES_DATA[var_to_read], var_to_read
            else:
                var_to_read = self.VARNAMES_DATA[var_to_read]

        if var_to_write is None:
            var_to_write = self.var_names_data_inv[var_to_read]

        with xr.open_dataset(filename, decode_timedelta=True) as ds:
            if not {"station", "time"}.issubset(ds.dims):  # pragma: no cover
                raise AttributeError("Missing dimensions")
            if "station_name" not in ds:  # pragma: no cover
                raise AttributeError("No variable station_name found")

            stats = []

            # get all station metadata values as numpy arrays, since xarray isel,
            # __getitem__, __getattr__ are slow... this can probably be solved
            # more elegantly
            meta_glob = {}
            for meta_key in self.META_KEYS:
                try:
                    meta_glob[meta_key] = ds[meta_key].values
                except KeyError:  # pragma: no cover
                    logger.warning(
                        f"No such metadata key in GHOST data file: {Path(filename).name}"
                    )

            for meta_key, to_unit in self.CONVERT_UNITS_META.items():
                from_unit = ds[meta_key].attrs["units"]

                meta_glob[meta_key] = convert_unit(
                    meta_glob[meta_key], from_unit=from_unit, to_unit=to_unit
                )

            tvals = ds["time"].values

            vardata = ds[var_to_read]  # DataArray
            varinfo = vardata.attrs

            # ToDo: it is important that station comes first since we use numpy
            # indexing below and not xarray.isel or similar, due to performance
            # issues. This may need to be updated in case of profile data.
            assert vardata.dims == ("station", "time")
            data_np = vardata.values

            # evaluate flags
            invalid = self._eval_flags(vardata, invalidate_flags, ds)

            for idx in ds.station.values:
                stat = {}
                meta = StationMetaData()
                meta["ts_type"] = self.TS_TYPE
                stat["time"] = tvals
                stat["meta"] = meta
                meta["var_info"] = {}

                for meta_key, vals in meta_glob.items():
                    meta[meta_key] = vals[idx]

                # vardata = subset[var_name]
                stat[var_to_write] = data_np[idx]

                meta["var_info"][var_to_write] = {}
                meta["var_info"][var_to_write].update(varinfo)

                # import flagdata (2D array with time and flag dimensions)
                # invalid = self._eval_flags(vardata, invalidate_flags)
                stat["data_flagged"] = {}
                stat["data_flagged"][var_to_write] = invalid[idx]
                stats.append(stat)

        return stats

    def _add_flags_var_to_compute(self, statlist_from_file, var_to_compute):
        for stat in statlist_from_file:
            for i, req in enumerate(self.AUX_REQUIRES[var_to_compute]):
                flags = stat["data_flagged"][req]
                if i == 0:
                    # pointer (safes computation time in case only one variable
                    # is required, i.e. the same flags can be used)
                    stat["data_flagged"][var_to_compute] = flags
                else:
                    logger.warning(
                        "THIS HAS NOT BEEN TESTED AND IS "
                        "SHOULD CURRENTLY NOT BE ABLE "
                        "TO BE REACHED."
                    )
                    current = stat["data_flagged"][var_to_compute].copy()
                    updated = np.logical_or(current, flags)
                    stat["data_flagged"][var_to_compute] = updated
        return statlist_from_file

    def compute_additional_vars(self, statlist_from_file, vars_to_compute):
        """
        Compute additional variables for all sites

        Parameters
        ----------
        statlist_from_file : list
            list of :class:`StationData` objects containing variable data that
            can be read from the data files.
        vars_to_compute : list
            list of variables to be computed from the variables contained in
            `statlist_from_file`.

        Returns
        -------
        statlist_from_file : list
            list of modified :class:`StationData` objects, containing computed
            variables in addition to the data that was contained in them
            initially
        vars_added : list
            list of variables that could be successfully added

        """
        vars_added = []
        for var in vars_to_compute:
            first_stat = statlist_from_file[0]
            can_compute = True
            requires = self.AUX_REQUIRES[var]
            for req in requires:
                if req not in first_stat:
                    can_compute = False
            if can_compute:
                # this will add the variable data to each station data in
                # statlist_from_file
                statlist_from_file = self.AUX_FUNS[var](statlist_from_file, var, *requires)
                statlist_from_file = self._add_flags_var_to_compute(statlist_from_file, var)

                if var not in vars_added:
                    vars_added.append(var)
        return (statlist_from_file, vars_added)

    def read(
        self,
        vars_to_retrieve=None,
        files=None,
        first_file=None,
        last_file=None,
        pattern=None,
        check_time=True,
        **kwargs,
    ):
        """Read data files into :class:`UngriddedData` object

        Parameters
        ----------
        vars_to_retrieve : list, optional
            list containing variables that are supposed to be read. If None,
            :attr:`DEFAULT_VARS` is used.
        files : list, optional
            list of files to be read. If None, then the file list is used that
            is returned on :func:`get_file_list`.
        first_file : int, optional
            index of first file in file list to read. If None, the first
            file in the list is used
        last_file : int, optional
            index of last file in list to read. If None, the last file
            in the list is used
         file_pattern : str, optional
            string pattern for file search (cf :func:`get_file_list`)

        Returns
        -------
        UngriddedData
            data object
        """
        if vars_to_retrieve is None:
            vars_to_retrieve = self.DEFAULT_VARS
        elif isinstance(vars_to_retrieve, str):
            vars_to_retrieve = [vars_to_retrieve]

        # make sure to use AeroCom variable names in output data
        vars_to_retrieve = varlist_aerocom(vars_to_retrieve)

        vars_to_read, vars_to_compute = self.check_vars_to_retrieve(vars_to_retrieve)

        if files is None:
            files = self.get_file_list(vars_to_read, pattern=pattern)
        elif isinstance(files, str):
            files = [files]

        if first_file is None:
            first_file = 0
        if last_file is None:
            last_file = len(files)

        files = files[first_file:last_file]

        data_obj = UngriddedData(num_points=1000000)

        meta_key = -1.0
        idx = 0

        # assign metadata object
        metadata = data_obj.metadata
        meta_idx = data_obj.meta_idx
        var_count_glob = -1
        rename = self.var_names_data_inv
        from tqdm import tqdm

        for i in tqdm(range(len(files))):
            _file = files[i]
            metafile = self.get_meta_filename(_file)
            var_to_read = metafile["var_name"]
            begin = metafile["start"]
            end = metafile["stop"]

            var_read = rename[var_to_read]
            stats = self.read_file(_file, var_to_read=var_to_read, var_to_write=var_read, **kwargs)

            stats, added = self.compute_additional_vars(stats, vars_to_compute)
            if len(stats) == 0:
                logger.info(
                    f"File {_file} does not contain any of the input variables {vars_to_retrieve}"
                )
            vars_avail = [var_read] + added
            vars_to_add = list(np.intersect1d(vars_to_retrieve, vars_avail))
            if len(vars_to_add) == 0:
                continue
            chunksize = 500000
            for stat in stats:
                meta_key += 1
                meta_idx[meta_key] = {}

                meta = stat["meta"]
                vi = meta["var_info"]

                meta["var_info"] = {}

                metadata[meta_key] = meta
                metadata[meta_key]["data_id"] = self.data_id
                # duplicate for now
                metadata[meta_key]["instrument_name"] = meta["measuring_instrument_name"]
                statname = metadata[meta_key]["station_name"]
                if "/" in statname:
                    statname = statname.replace("/", "-")
                metadata[meta_key]["station_name"] = statname

                times = stat["time"].astype("datetime64[s]")
                timenums = np.float64(times)

                if check_time and (begin > times[0] or end < times[-1]):
                    raise ValueError("Something seems to be off with time dimension...")

                num_vars = len(vars_to_add)
                num_times = len(times)

                totnum = num_times * num_vars

                # check if size of data object needs to be extended
                if (idx + totnum) >= data_obj._ROWNO:
                    # if totnum < data_obj._CHUNKSIZE, then the latter is used
                    data_obj.add_chunk(chunksize)

                for j, var_to_write in enumerate(vars_to_add):
                    values = stat[var_to_write]

                    start = idx + j * num_times
                    stop = start + num_times

                    if var_to_write not in data_obj.var_idx:
                        var_count_glob += 1
                        var_idx = var_count_glob
                        data_obj.var_idx[var_to_write] = var_idx
                    else:
                        var_idx = data_obj.var_idx[var_to_write]

                    meta["var_info"][var_to_write] = vi[var_to_write]
                    # write common meta info for this station (data lon, lat and
                    # altitude are set to station locations)
                    data_obj._data[start:stop, data_obj._LATINDEX] = meta["latitude"]
                    data_obj._data[start:stop, data_obj._LONINDEX] = meta["longitude"]
                    data_obj._data[start:stop, data_obj._ALTITUDEINDEX] = meta["altitude"]
                    data_obj._data[start:stop, data_obj._METADATAKEYINDEX] = meta_key

                    # write data to data object
                    data_obj._data[start:stop, data_obj._TIMEINDEX] = timenums

                    data_obj._data[start:stop, data_obj._DATAINDEX] = values

                    # add invalid measurements
                    invalid = stat["data_flagged"][var_to_write]
                    data_obj._data[start:stop, data_obj._DATAFLAGINDEX] = invalid

                    data_obj._data[start:stop, data_obj._VARINDEX] = var_idx

                    meta_idx[meta_key][var_to_write] = np.arange(start, stop)

                idx += totnum

        data_obj._data = data_obj._data[:idx]
        data_obj._check_index()
        return data_obj



from __future__ import annotations

import logging
import re
import time
import warnings
from pathlib import Path
from reprlib import repr

import numpy as np
import xarray as xr

from pyaerocom import ColocatedData
from pyaerocom.aeroval._processing_base import ProcessingEngine
from pyaerocom.aeroval.coldatatojson_helpers import (
    _select_period_season_coldata,
    init_regions_web,
)
from pyaerocom.aeroval.fairmode_statistics import SPECIES, FairmodeStatistics
from pyaerocom.exceptions import DataCoverageError, UnknownRegion
from pyaerocom.io.cams2_83.models import ModelName
from pyaerocom.units.datetime import TsType

logger = logging.getLogger(__name__)
warnings.filterwarnings("ignore")


class CAMS2_83_Engine(ProcessingEngine):
    MEDIANSCORE_SPECIES = [
        "concno2",
        "concco",
        "conco3",
        "concso2",
        "concpm10",
        "concpm25",
    ]

    def run(
        self, files: list[list[str | Path]], var_list: list
    ) -> None:  # type:ignore[override]
        logger.info(f"Processing: {repr(files)}")
        coldata = [ColocatedData(data=file) for file in files]
        coldata, persistence_cols, found_vars, found_persistence = self._sort_coldata(
            coldata
        )
        start = time.time()

        if var_list is None:
            var_list_2 = list(found_vars)
            if not var_list_2:
                logging.warning(
                    f"No variables found in colocated data var_list={var_list}, found_vars={found_vars}"
                )
                return
        elif var_list == ["conco3"] or (len(var_list) > 1 and "conco3" in var_list):
            var_list_2 = list(var_list)
            var_list_2.append("conco3mda8")
        else:
            var_list_2 = list(var_list)

        logger.info(f"Running CAMS2_83_Engine.run with var_list_2 {var_list_2}")
        for var in var_list_2:
            if var not in found_vars:
                logger.warning(f"{var} not found in coldata, skipping")
                continue
            logger.info(f"Processing Component: {var}")
            if found_persistence:
                self.process_coldata(coldata[var], persistence_cols[var], var)
            else:
                self.process_coldata(coldata[var], [], var)

        logger.info(f"Time for weird plot: {time.time() - start} sec")

    def process_coldata(
        self,
        coldata: list[ColocatedData],
        persistence_coldata: list[ColocatedData],
        var_name: str,
    ) -> None:
        use_weights = self.cfg.statistics_opts.weighted_stats
        forecast_days = self.cfg.statistics_opts.forecast_days
        periods = self.cfg.time_cfg.periods
        min_num_obs = self.cfg.colocation_opts.min_num_obs

        # use_fairmode = self.cfg.statistics_opts.use_fairmode
        use_fairmode = self.cfg.cams2_83_cfg.use_cams2_83_fairmode
        calc_forecast_target = False

        calc_medianscores = True if var_name in self.MEDIANSCORE_SPECIES else False

        if use_fairmode:
            fairmode_statistics = FairmodeStatistics()

        if use_fairmode and len(persistence_coldata) > 0 and var_name in SPECIES:
            persistence_coldata = persistence_coldata[0]
            calc_forecast_target = True

            if SPECIES[var_name]["freq"] != TsType("hourly"):
                persistence_coldata = persistence_coldata.resample_time(
                    SPECIES[var_name]["freq"],
                    settings_from_meta=True,
                    min_num_obs=min_num_obs,
                )

        if "var_name_input" in coldata[0].metadata:
            obs_var = coldata[0].metadata["var_name_input"][0]
            model_var = coldata[0].metadata["var_name_input"][1]
        else:
            obs_var = model_var = "UNDEFINED"

        # for the MOS/ENS evaluation experiment the models are just strings
        # we do not want them added to the ModelName class
        # so we need a bunch of ugly special cases here
        modelname = coldata[0].model_name.split("-")[2]
        if modelname == "ENS" or modelname == "MOS":
            model = modelname
        else:
            model = ModelName[modelname]
        vert_code = coldata[0].get_meta_item("vert_code")
        obs_name = coldata[0].obs_name
        if modelname == "ENS" or modelname == "MOS":  # MOS/ENS evaluation special case
            mcfg = self.cfg.model_cfg.get_entry(modelname)
        else:
            mcfg = self.cfg.model_cfg.get_entry(model.name)
        var_name_web = mcfg.get_varname_web(model_var, obs_var)
        seasons = self.cfg.time_cfg.get_seasons()

        regions_how = "country"
        use_country = True
        for i in range(forecast_days):
            coldata[i].data["season"] = coldata[i].data.time.dt.season
            (regborders, regs, regnames) = init_regions_web(coldata[i], regions_how)

        if calc_forecast_target:
            persistence_coldata.data["season"] = persistence_coldata.data.time.dt.season
            (regborders, regs, regnames) = init_regions_web(
                persistence_coldata, regions_how
            )
            # results_mqi = {}
        results = {}
        results_fairmode = {}

        for regid, regname in regnames.items():
            results[regname] = {}
            results_fairmode[regname] = {}
            logger.info(f"Creating subset for {regname}")
            try:
                subset_region = [
                    col.filter_region(regid, check_country_meta=use_country)
                    for col in coldata
                ]
                if calc_forecast_target:
                    persistence_subset_region = persistence_coldata.filter_region(
                        regid, check_country_meta=use_country
                    )
                    # results_mqi[regname] = {}
            except (DataCoverageError, UnknownRegion) as e:
                logger.info(
                    f"Skipping forecast plot for {regname} due to error {str(e)}"
                )
                continue
            for per in periods:
                for season in seasons:
                    perstr = f"{per}-{season}"

                    stats_list: dict[str, list[float]] = dict(
                        rms=[], R=[], nmb=[], mnmb=[], fge=[]
                    )
                    logger.info(f"Making subset for {regid}, {per} and {season}")
                    if season not in coldata[0].data["season"].data and season != "all":
                        logger.info(
                            f"Season {season} is not available for {per} and will be skipped"
                        )
                        continue

                    try:
                        subset = [
                            _select_period_season_coldata(
                                col, per, season, use_meteorological_seasons=True
                            )
                            for col in subset_region
                        ]
                    except (DataCoverageError, UnknownRegion) as e:
                        logger.info(f"Skipping forecast plot due to error {str(e)}")
                        continue
                    if calc_medianscores:
                        for forecast_hour in range(24 * forecast_days):
                            logger.debug(
                                f"Calculating statistics for hour {forecast_hour}"
                            )
                            leap, hour = divmod(forecast_hour, 24)
                            ds = subset[leap]
                            ds = ds.data.sel(time=(ds.time.dt.hour == hour))
                            start = time.time()
                            stats = self._get_median_stats_point_vec(ds, use_weights)
                            logger.debug(time.time() - start)
                            for key in stats_list:
                                stats_list[key].append(stats[key])

                    if use_fairmode and var_name in SPECIES:

                        fairmode_subset = subset[0]
                        if SPECIES[var_name]["freq"] != TsType("hourly"):
                            fairmode_subset = fairmode_subset.resample_time(
                                SPECIES[var_name]["freq"],
                                settings_from_meta=True,
                                min_num_obs=min_num_obs
                            )

                        results_fairmode[f"{regname}"][f"{perstr}"] = (
                            fairmode_statistics.fairmode_statistics(
                                fairmode_subset, var_name
                            )
                        )

                        if calc_forecast_target:
                            
                            results_mqi = []
                            for day in range(forecast_days):
                                ds = subset[day]
                                ds_p = persistence_subset_region

                                
                                mqi_results = self._calc_forecast_target_MQI_vectorized(
                                    ds, ds_p, var_name, day, min_num_obs
                                )

                                results_mqi.append(mqi_results)

                               

                            for station in results_fairmode[f"{regname}"][f"{perstr}"]:
                                results_fairmode[f"{regname}"][f"{perstr}"][station][
                                    "rms"
                                ] = []
                                results_fairmode[f"{regname}"][f"{perstr}"][station][
                                    "beta_mqi"
                                ] = []
                                results_fairmode[f"{regname}"][f"{perstr}"][station][
                                    "sign"
                                ] = []

                                results_fairmode[f"{regname}"][f"{perstr}"][station][
                                    "persistence_model"
                                ] = True
                                for day in range(forecast_days):
                                    mqi_p = (
                                        results_mqi[day][station]
                                        if station in results_mqi[day]
                                        else [np.nan, np.nan, np.nan]
                                    )
                                    results_fairmode[f"{regname}"][f"{perstr}"][
                                        station
                                    ]["rms"].append(mqi_p[0])
                                    results_fairmode[f"{regname}"][f"{perstr}"][
                                        station
                                    ]["beta_mqi"].append(mqi_p[1])
                                    results_fairmode[f"{regname}"][f"{perstr}"][
                                        station
                                    ]["sign"].append(mqi_p[2])

                    out_dirs = self.cfg.path_manager.get_json_output_dirs(
                        True
                    )  # noqa: F841

                    results[f"{regname}"][f"{perstr}"] = stats_list

            if calc_medianscores:
                self.exp_output.add_forecast_entry(
                    results[regname],
                    regname,
                    obs_name,
                    var_name_web,
                    vert_code,
                    (
                        modelname
                        if (modelname == "ENS" or modelname == "MOS")
                        else model.name
                    ),  # MOS/ENS evaluation special case
                    model_var,
                )

        if use_fairmode and var_name in SPECIES:
            fairmode_statistics.save_fairmode_stats(
                self.exp_output,
                results_fairmode,
                obs_name,
                var_name_web,
                vert_code,
                (
                    modelname
                    if (modelname == "ENS" or modelname == "MOS")
                    else model.name
                ),  # MOS/ENS evaluation special case
                model_var,
            )

    def _get_median_stats_point(
        self, data: xr.DataArray, use_weights: bool
    ) -> dict[str, float]:
        stats_list: dict[str, list[float]] = dict(rms=[], R=[], nmb=[], mnmb=[], fge=[])
        station_list = data.station_name.data
        for station in station_list:
            d = data.sel(station_name=[station])
            arr = ColocatedData(data=d)
            stats = arr.calc_statistics(use_area_weights=use_weights)
            for key in stats_list.keys():
                stats_list[key].append(stats[key])
        median_stats = {}
        for key in stats_list.keys():
            median_stats[key] = np.nanmedian(np.array(stats_list[key]))

        return median_stats

    def _get_median_stats_point_vec(
        self, data: xr.DataArray, use_weights: bool
    ) -> dict[str, float]:
        stats_list: dict[str, float] = dict(rms=0.0, R=0.0, nmb=0.0, mnmb=0.0, fge=0.0)

        obsvals = data.data[0]
        modvals = data.data[1]

        mask = ~np.isnan(obsvals) * ~np.isnan(modvals)

        diff = modvals - obsvals
        diffsquare = diff**2
        sum_obs = np.sum(obsvals, axis=0, where=mask)
        sum_diff = np.sum(diff, axis=0, where=mask)
        sum_vals = obsvals + modvals

        tmp = diff / sum_vals

        nmb = np.where(sum_obs == 0, np.nan, sum_diff / sum_obs)

        mnmb = 2.0 * np.mean(tmp, axis=0, where=mask)
        fge = 2.0 * np.mean(np.abs(tmp), axis=0, where=mask)
        rms = np.sqrt(np.mean(diffsquare, axis=0, where=mask))

        R = self._pearson_R_vec(obsvals, modvals)

        stats_list["rms"] = np.nanmedian(rms)
        stats_list["R"] = np.nanmedian(R)
        stats_list["nmb"] = np.nanmedian(nmb)
        stats_list["mnmb"] = np.nanmedian(mnmb)
        stats_list["fge"] = np.nanmedian(fge)

        return stats_list

    def _pearson_R_vec(self, x: np.ndarray, y: np.ndarray) -> np.ndarray:
        return FairmodeStatistics.pearson_R(x, y)

    def _calc_forecast_target_MQI_vectorized(
        self,
        coldata: ColocatedData,
        persistence_coldata: ColocatedData,
        var_name: str,
        forecast_day: int,
        min_num_obs: dict,
    ) -> dict[str, float]:

        results = {}

        # Resampling of time for all other variables than NO2
        if SPECIES[var_name]["freq"] != TsType("hourly"):
            coldata = coldata.resample_time(
                SPECIES[var_name]["freq"],
                settings_from_meta=True,
                min_num_obs=min_num_obs
            )

        # Creation of mask of shared stations between normal data and persistence data
        # stations = persistence_coldata.data.station_name.values
        station_mask = np.intersect1d(
            persistence_coldata.data.station_name.values,
            coldata.data.station_name.values,
            return_indices=True,
        )
        assert np.all(
            persistence_coldata.data.station_name.values[station_mask[1]]
            == coldata.data.station_name.values[station_mask[2]]
        )

        # Gets times. Moves percistence time forward, indicating that the obs on day N is the perstence model on day N+1 (or rather N+forecast_day+1)
        data_time = coldata.time.values
        p_time = persistence_coldata.time.values + np.timedelta64(
            24 * (forecast_day + 1), "h"
        )  # This needs to be checked if it is correct for forecast days > 0

        # Masks the time, so that only dates which has a valid persistence model and data are used
        time_mask = np.intersect1d(p_time, data_time, return_indices=True)

        # Fetching of masked data
        obs_vals = coldata.data.data[0, :, station_mask[2]][:, time_mask[2]]
        mod_vals = coldata.data.data[1, :, station_mask[2]][:, time_mask[2]]
        p_mod_vals = persistence_coldata.data.data[0, :, station_mask[1]][
            :, time_mask[1]
        ]  # persistence model

        # Gets a NaN mask
        mask = ~np.isnan(obs_vals) * ~np.isnan(mod_vals) * ~np.isnan(p_mod_vals)

        # Sanity Check
        assert np.all(p_mod_vals.shape == obs_vals.shape)

        sign = self._target_plot_sign(
            obsvals=obs_vals, modvals=mod_vals, mask=mask, var_name=var_name
        )

        # Calculation of MQI
        factor = SPECIES[var_name]["alpha"] ** 2 * SPECIES[var_name]["RV"] ** 2
        uncertainty_p_obs = SPECIES[var_name]["UrRV"] * np.sqrt(
            (1 - SPECIES[var_name]["alpha"] ** 2) * p_mod_vals**2 + factor
        )

        p_diff_vals = np.maximum(
            np.abs(obs_vals - p_mod_vals - uncertainty_p_obs),
            np.abs(obs_vals - p_mod_vals + uncertainty_p_obs),
        )

        rmse_m = np.sqrt(np.nanmean((mod_vals - obs_vals) ** 2, axis=1, where=mask))
        rmse_p = np.sqrt(np.nanmean((p_diff_vals) ** 2, axis=1, where=mask))

        mqi = rmse_m / rmse_p

        results = {
            str(station_mask[0][i]): [rmse_p[i], mqi[i], sign[i]]
            for i in range(len(mqi))
        }

        return results

    def _target_plot_sign(
        self,
        obsvals: np.ndarray,
        modvals: np.ndarray,
        mask: np.ndarray,
        var_name: str,
    ) -> np.ndarray:

        threshold = SPECIES[var_name]["RV"]
        false_alarms = np.sum(
            np.logical_and(modvals > threshold, obsvals <= threshold, where=mask),
            axis=1,
            where=mask,
        )
        missed_alarms = np.sum(
            np.logical_and(modvals <= threshold, obsvals > threshold, where=mask),
            axis=1,
            where=mask,
        )
        sign = np.where(
            false_alarms <= missed_alarms, -1.0, 1.0
        )  
        return sign

    def _sort_coldata(
        self, coldata: list[ColocatedData]
    ) -> tuple[
        dict[str, list[ColocatedData]], dict[str, list[ColocatedData]], set[str], bool
    ]:
        col_dict = dict()
        persistence_dict = dict()

        persistence_var_list = []
        var_list = []

        found_persistence = False

        for col in coldata:
            obs_var = col.metadata["var_name_input"][0]

            if "persistence" in col.model_name:
                found_persistence = True
                if obs_var in persistence_dict:
                    persistence_dict[obs_var].append(col)
                else:
                    persistence_dict[obs_var] = [col]
                    persistence_var_list.append(obs_var)
            else:
                if obs_var in col_dict:
                    col_dict[obs_var].append(col)
                else:
                    col_dict[obs_var] = [col]
                    var_list.append(obs_var)
        if found_persistence:
            logger.info(f"persistence model has been found for {persistence_var_list}")
            assert set(sorted(var_list)) == set(sorted(persistence_var_list))
        for var, cols in col_dict.items():
            col_dict[var] = sorted(cols, key=lambda x: self._get_day(x.model_name))

        return col_dict, persistence_dict, set(var_list), found_persistence

    def _get_day(self, model_name: str) -> int:
        return int(re.search(".*day([0-3]).*", model_name).group(1))



import numpy as np
import pytest

from pyaerocom.stats.implementations import stat_mb, stat_R, stat_R_kendall
from pyaerocom.stats.stats import calculate_statistics


def test_calc_stats_exceptions():
    with pytest.raises(ValueError):
        calculate_statistics(
            np.asarray([[1, 2, 3, 4], [1, 2, 3, 4]]), np.asarray([[1, 2, 3, 4], [1, 2, 3, 4]])
        )

    with pytest.raises(ValueError):
        calculate_statistics([1, 2, 3, 4], [1, 2, 3, 4], weights=[1, 2, 3])

    with pytest.raises(ValueError):
        calculate_statistics([1, 2, 3], [1, 2])

    with pytest.raises(ValueError):
        calculate_statistics([1, 2, 3], [1, 2, 3], weights=[1, 2])


@pytest.mark.parametrize(
    "data,ref_data,statistics,expected_keys",
    (
        pytest.param(
            [1, 2, 3, 4], [1, 2, 3, 4], {}, ["totnum", "weighted", "num_valid"], id="empty"
        ),
        pytest.param(
            [1, 2, 3, 4],
            [1, 2, 3, 4],
            {"R": stat_R, "R_kendall": stat_R_kendall, "mb": stat_mb},
            ["totnum", "weighted", "num_valid", "R", "R_kendall", "mb"],
            id="custom1",
        ),
        pytest.param(
            [1, 2, 3, 4],
            [1, 2, 3, 4],
            None,
            [
                "totnum",
                "weighted",
                "num_valid",
                "R",
                "R_kendall",
                "R_spearman",
                "fge",
                "mab",
                "mab",
                "mb",
                "mnmb",
                "nmb",
                "rms",
                "data_mean",
                "data_std",
                "refdata_mean",
                "refdata_std",
            ],
            id="stats-none",
        ),
    ),
)
def test_calc_stats_keys(data, ref_data, statistics, expected_keys):
    stats = calculate_statistics(data, ref_data, statistics)

    assert isinstance(stats, dict)

    for k in expected_keys:
        assert k in stats.keys()

    for k in stats.keys():
        assert k in expected_keys


def test_custom_stats():
    stats = calculate_statistics(
        [1, 2, 3, 4],
        [1, 2, 3, 4],
        {"test_statistic1": lambda x, y, w: 5, "test_statistic2": lambda x, y, w: 3},
    )

    assert "test_statistic1" in stats.keys()
    assert "test_statistic2" in stats.keys()
    assert stats["test_statistic1"] == 5
    assert stats["test_statistic2"] == 3


perfect_stats_num1_mean1 = {
    "totnum": 1,
    "num_valid": 1,
    "refdata_mean": 1,
    "refdata_std": 0,
    "data_mean": 1,
    "data_std": 0,
    "weighted": False,
    "rms": 0,
    "nmb": 0,
    "mnmb": 0,
    "mb": 0,
    "mab": 0,
    "fge": 0,
    "R": np.nan,
    "R_kendall": np.nan,
    "R_spearman": np.nan,
}
perfect_stats_num2_mean1 = perfect_stats_num1_mean1.copy()
perfect_stats_num2_mean1.update(totnum=2)

num_fakedata = 1000
zero_signal = np.zeros(num_fakedata)
noise = np.random.normal(loc=0, scale=0.01, size=num_fakedata)

idx = np.linspace(0, 2 * np.pi, num_fakedata)
sin_signal = np.sin(idx)


@pytest.mark.parametrize(
    "data,ref_data,expected",
    [
        ([1], [1], perfect_stats_num1_mean1),
        ([1, np.nan], [1, np.nan], perfect_stats_num2_mean1),
        (
            zero_signal,
            zero_signal,
            {
                "totnum": 1000.0,
                "num_valid": 1000.0,
                "refdata_mean": 0.0,
                "refdata_std": 0,
                "data_mean": 0.0,
                "data_std": 0,
                "weighted": False,
                "rms": 0.0,
                "R": np.nan,
                "R_spearman": np.nan,
                "R_kendall": np.nan,
                "nmb": 0,
                "mb": 0,
                "mab": 0,
                "mnmb": np.nan,
                "fge": np.nan,
            },
        ),
        (
            zero_signal,
            noise,
            {
                "totnum": 1000.0,
                "num_valid": 1000.0,
                "refdata_mean": 0.0,
                "refdata_std": 0,
                "data_mean": 0.0,
                "data_std": 0,
                "weighted": False,
                "rms": 0.0,
                "R": np.nan,
                "R_spearman": np.nan,
                "R_kendall": np.nan,
                "nmb": -1,
                "mb": 0,
                "mab": 0,
                "mnmb": -2,
                "fge": 2,
            },
        ),
        (
            sin_signal,
            sin_signal,
            {
                "totnum": 1000.0,
                "num_valid": 1000.0,
                "refdata_mean": 0.0,
                "refdata_std": 0.71,
                "data_mean": 0.0,
                "data_std": 0.71,
                "weighted": False,
                "rms": 0.0,
                "R": 1.0,
                "R_spearman": 1.0,
                "R_kendall": 1.0,
                "nmb": 0,
                "mb": 0,
                "mab": 0,
                "mnmb": np.nan,
                "fge": np.nan,
            },
        ),
    ],
)
def test_calc_statistics(data, ref_data, expected):
    stats = calculate_statistics(data, ref_data)
    assert isinstance(stats, dict)
    assert len(stats) == len(expected)
    for key, val in expected.items():
        assert key in stats
        assert stats[key] == pytest.approx(val, abs=0.02, rel=0.01, nan_ok=True)


@pytest.mark.parametrize(
    "data,ref_data,drop",
    (
        pytest.param(tuple(), tuple(), tuple(), id="empty-data"),
        pytest.param(tuple(), tuple(), ("R", "R_kendall"), id="empty-data-with-drop"),
        pytest.param((1, 2, 3, 4), (1, 2, 3, 4), ("R", "R_kendall"), id="small-data"),
        (
            (1,),
            (2,),
            (
                "nmb",
                "mnmb",
                "mb",
                "mab",
                "R",
                "R_spearman",
                "R_kendall",
                "fge",
                "nrms",
                "rms",
                "data_mean",
                "data_std",
                "weighted",
            ),
        ),
    ),
)
def test_calc_statistics_drop_stats(data, ref_data, drop):
    stats = calculate_statistics(data, ref_data, drop_stats=drop)

    assert all([x not in stats.keys() for x in drop])


def test_fairmode_R_in_expected_range():
    # This test is meant to test for a specific issue where
    # using fairmode=true would result in errors due to the calculated
    # R values being slightly out of range -1<=R<=1. The error is caused
    # because fairmode_stats() expects R values exactly in the range, but
    # values may be outside due to floating point shenanigans.
    # https://xkcd.com/217/
    # As of the time of writing this test, the data used for this test is
    # known to result in a failing test, IF *no* rounding is applied in
    # `stats.stats._prepare_statistics().`
    # https://github.com/metno/pyaerocom/pull/1142
    data1 = np.asarray([17.2073523845202, 15.543783901903302, np.nan])
    data2 = np.asarray([15.046913580246915, 18.462202650290884, np.nan])

    stats = calculate_statistics(data1, data2)

    assert -1 <= stats["R"] <= 1



import numpy as np
import pytest
import xarray as xr

from pyaerocom.colocation.colocated_data import ColocatedData
from pyaerocom.stats.mda8.mda8 import (
    _calc_mda8,
    _daily_max,
    _rolling_average_8hr,
    mda8_colocated_data,
)


@pytest.fixture
def test_data(time, values) -> xr.DataArray:
    return xr.DataArray(
        [[[x] for x in values]],
        dims=["data_source", "time", "station_name"],
        coords={"time": time},
    )


@pytest.mark.parametrize(
    "time,values,exp_mda8",
    (
        pytest.param(
            xr.date_range(start="2024-01-01 01:00", periods=49, freq="1h"),
            [0.0] * 49,
            [0, 0, np.nan],
            id="zeros",
        ),
        pytest.param(
            xr.date_range(start="2024-01-01 01:00", periods=49, freq="1h"),
            np.linspace(start=1, stop=49, num=49),
            [20.5, 44.5, np.nan],
            id="incrementing-by-1",
        ),
        pytest.param(
            xr.date_range(start="2024-01-01 13:00:00", periods=49, freq="1h"),
            np.concatenate(  # Sequence 1-7, 3x nan, 11-14, 3x nan, 28-49
                (
                    np.arange(start=1, stop=8),
                    [np.nan] * 3,
                    np.arange(start=11, stop=25),
                    [np.nan] * 3,
                    np.arange(start=28, stop=50),
                )
            ),
            [np.nan] * 3,
            id="with-nans",
        ),
        # https://github.com/metno/pyaerocom/issues/1323
        pytest.param(
            xr.date_range(start="2024-01-01 06:00:00", periods=30, freq="1h"),
            np.arange(30, dtype=float),
            [np.nan, np.nan],
            id="#1323",
        ),
    ),
)
def test_calc_mda8(test_data, exp_mda8):
    mda8 = _calc_mda8(test_data)

    assert mda8.shape[1] == len(exp_mda8)

    assert all(mda8[0, :, 0] == pytest.approx(exp_mda8, abs=0, nan_ok=True))


def test_calc_mda8_with_gap():
    arr1 = xr.DataArray(
        [[[x] for x in np.linspace(start=1, stop=50, num=50)]],
        dims=["data_source", "time", "station_name"],
        coords={"time": xr.date_range(start="2024-01-01 01:00", periods=50, freq="1h")},
    )

    arr2 = xr.DataArray(
        [[[x] for x in np.linspace(start=1, stop=50, num=50)]],
        dims=["data_source", "time", "station_name"],
        coords={"time": xr.date_range(start="2024-01-04 01:00", periods=50, freq="1h")},
    )

    arr = xr.concat((arr1, arr2), dim="time")

    mda8 = _calc_mda8(arr)

    assert mda8.shape == (1, 6, 1)
    pytest.approx(mda8[0, :, 0], [20.5, 44.5, np.nan, 41.25, 44.5, np.nan], abs=10 * 10**-5)


@pytest.mark.parametrize("coldataset", ("fake_3d_hr",))
def test_coldata_to_mda8(coldata):
    mda8 = mda8_colocated_data(coldata, obs_var="vmro3mda8", mod_var="vmro3mda8")

    assert isinstance(mda8, ColocatedData)
    assert mda8.metadata["ts_type"] == "daily"
    assert mda8.metadata["var_name"] == ["vmro3mda8", "vmro3mda8"]
    assert mda8.shape == (2, 8, 1)

    assert mda8.data.values[0, :, 0] == pytest.approx(
        [np.nan, np.nan, 1.18741556, 1.18777241, 1.18869106, 1.18879322, 1.18807846, 1.18700801],
        abs=10**-5,
        nan_ok=True,
    )

    assert mda8.data.values[1, :, 0] == pytest.approx(
        [
            1.57327333,
            1.28884431,
            1.28741556,
            1.28777241,
            1.28869106,
            1.28879322,
            1.28807846,
            1.28700801,
        ],
        abs=10**-5,
        nan_ok=True,
    )


@pytest.mark.parametrize(
    "time,values,exp_ravg",
    (
        (
            xr.date_range(start="2024-01-01 01:00", periods=48, freq="1h"),
            [0.0, 1, 2, 3, 4, 5, 6, 7] * 6,
            [np.nan] * 5 + [2.5, 3] + [3.5] * 41,
        ),
        (
            xr.date_range(start="2024-01-01 01:00", periods=24, freq="1h"),
            [np.nan, 1, 2, 3, 4, 5, 6, 7] * 3,
            [np.nan] * 6 + [3.5] + [4] * 17,
        ),
        (
            xr.date_range(start="2024-01-01 01:00", periods=24, freq="1h"),
            [np.nan, np.nan, 2, 3, 4, 5, 6, 7] * 3,
            [np.nan] * 7 + [4.5] * 17,
        ),
        (
            xr.date_range(start="2024-01-01 01:00", periods=24, freq="1h"),
            [np.nan, np.nan, np.nan, 3, 4, 5, 6, 7] * 3,
            [np.nan] * 24,
        ),
    ),
)
def test_rollingaverage(test_data, exp_ravg):
    """
    Test of rolling average calculation
    """
    ravg = _rolling_average_8hr(test_data)

    assert all(ravg[0, :, 0] == pytest.approx(exp_ravg, abs=0, nan_ok=True))


def test_rollingaverage_label():
    """
    Checks that the labels of rolling average is correct (we want it to be labeled based
    on the "right-most" interval, ie. latest measurement in the interval). This is currently
    the case in xarray but not greatly documented, so this test checks for that.

    https://github.com/metno/pyaerocom/issues/1323
    """
    data = xr.DataArray(
        [[[float(x)] for x in range(24)]],
        dims=["data_source", "time", "station_name"],
        coords={"time": xr.date_range(start="2024-01-01 00:00", periods=24, freq="1h")},
    )

    ravg = _rolling_average_8hr(data)

    assert ravg["time"].isel(time=0) == np.datetime64("2024-01-01 00:00")
    assert ravg["time"].isel(time=23) == np.datetime64("2024-01-01 23:00")


@pytest.mark.parametrize(
    "time,values,exp_daily_max",
    (
        (
            xr.date_range(start="2024-01-01 01:00", periods=48, freq="1h"),
            np.linspace(1, 48, num=48),
            [24, 48],
        ),
        (
            xr.date_range(start="2024-01-01 01:00", periods=24, freq="1h"),
            [np.nan] * 6 + list(range(7, 25)),
            [24],
        ),
        (
            xr.date_range(start="2024-01-01 01:00", periods=24, freq="1h"),
            [np.nan] * 7 + list(range(8, 25)),
            [np.nan],
        ),
    ),
)
def test_daily_max(test_data, exp_daily_max):
    """
    Tests the daily max calculation. Note that "the first calculation period
    for any one day will be the period from 17:00 on the previous day to 01:00
    on that day; the last calculation period for any one day will be the period
    from 16:00 to 24:00 on that day
    """
    dmax = _daily_max(test_data)

    assert all(dmax[0, :, 0] == pytest.approx(exp_daily_max, abs=0, nan_ok=True))



import pytest

from pyaerocom import geodesy
from tests.conftest import etopo1_unavail

TEST_LAT = 50.8
TEST_LON = 9


@pytest.mark.parametrize(
    "coords,countries,country_code",
    [
        ((52, 12), ["Germany"], ["DE"]),
        ([(46.1956, 6.21125), (55.398, 10.3669)], ["France", "Denmark"], ["FR", "DK"]),
    ],
)
def test_get_country_info_coords(coords, countries, country_code):
    for i, res in enumerate(geodesy.get_country_info_coords(coords)):
        assert isinstance(res, dict)
        assert "country" in res
        assert res["country"] == countries[i]
        assert res["country_code"] == country_code[i]


def test_haversine():
    assert geodesy.haversine(0, 15, 0, 16) == pytest.approx(111.2, abs=0.1)


def test_is_within_radius_km():
    assert geodesy.is_within_radius_km(0, 15, 0, 16, 1000, 111.2)


@pytest.mark.xfail(
    reason="fails in CI when geonum can not download 'srtm' data",
    raises=FileNotFoundError,
    strict=False,
)
def test_srtm_altitude():
    assert geodesy.get_topo_altitude(TEST_LAT, TEST_LON) == pytest.approx(207)


@etopo1_unavail
def test_etopo_altitude():
    alt = geodesy.get_topo_altitude(TEST_LAT, TEST_LON, topo_dataset="etopo1")
    assert alt == pytest.approx(217)



from collections.abc import Mapping

import numpy as np

from pyaerocom.stats.data_filters import FilterByLimit, FilterNaN
from pyaerocom.stats.implementations import (
    stat_fge,
    stat_mab,
    stat_mb,
    stat_mnmb,
    stat_nmb,
    stat_R,
    stat_R_kendall,
    stat_R_spearman,
    stat_rms,
)
from pyaerocom.stats.stat_filters import FilterDropStats
from pyaerocom.stats.types import DataFilter, StatisticsCalculator, StatisticsFilter, StatsDict


def _filter_data(
    data: np.ndarray,
    ref_data: np.ndarray,
    weights: np.ndarray,
    filters=list[DataFilter] | None,
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Filter data arrays using provided filter functions.

    Parameters
    ----------
    data : np.ndarray
    ref_data : np.ndarray
    weights : np.ndarray | None
    filters : list[DataFilter] | None

    Returns
    -------
    Tuple
        tuple consisting of the modified data, ref_data and optional weights array.

    Raises:
    -------
    IndexError:
        If the length of a produced mask does not match the length of the data.
    """
    if filters is not None:
        for f in filters:
            mask = f(data, ref_data, weights)
            if len(mask) != len(data):
                raise IndexError(
                    f"Filter mask length ({len(mask)}) does not match length of data ({len(data)})."
                )
            if weights is None:
                data, ref_data, weights = data[mask], ref_data[mask], None
            else:
                data, ref_data, weights = data[mask], ref_data[mask], weights[mask]

    return (data, ref_data, weights)


def _prepare_statistics(
    data: np.ndarray,
    ref_data: np.ndarray,
    weights: np.ndarray,
    statistics=dict[str, StatisticsCalculator],
    min_numvalid: int = 1,
) -> StatsDict:
    """Calculates configured statistics for two one-dimensional data arrays.

    Parameters
    ----------
    data : np.ndarray
    ref_data : np.ndarray
    weights : np.ndarray | None
        Optional list of weights which will be used for weighted statistics.
    statistics : dict[str, StatisticsCalculator]
        Mapping between names and functions that calculate individual statistics.
    min_numvalid: int
        Minmium number of data points required. If data is lower than this threshold,
        stats will be set to NaN.

    Returns
    -------
    StatsDict
        Dictionary with the calculated statistics. The structure of the return dictionary mirrors
        the statistics parameter.

    Calculates the statistics configured by statistics returning them as a StatsDict.
    """
    result = dict()
    for name, stat in statistics.items():
        if len(data) >= min_numvalid:
            # Rounding to ensure values aren't out of bounds due to floating point
            # arithmetic and to reduce json file size.
            result[name] = np.round(stat(data, ref_data, weights), decimals=6)
        else:
            result[name] = np.nan

    return result


def _filter_stats(
    stats: StatsDict, filters: list[StatisticsFilter] | None
) -> dict[str, np.float64]:
    """Filter a StatsDict

    Filters a StatsDict, dropping values or statistics.

    Parameters
    ----------
    stats : StatsDict
        The StatsDict to be filtered.
    filters
        A list of filter callables which are applied to the dict. The functions
        should take a dict and return the modified dict.

    Returns
    -------
    StatsDict
        The filtered StatsDict.
    """
    if filters is not None:
        for f in filters:
            stats = f(stats)

    return stats


def _get_default_statistic_config() -> dict[str, StatisticsCalculator]:
    """
    Returns a base configuration dictionary to be used with `calculate_statistics`
    which calculates all implemented statistics. Can be used as a starting
    point for adding additional stats using `dict.update()`
    """
    return {
        "refdata_mean": lambda x, y, w: np.nanmean(y),
        "refdata_std": lambda x, y, w: np.nanstd(y),
        "data_mean": lambda x, y, w: np.nanmean(x),
        "data_std": lambda x, y, w: np.nanstd(x),
        "rms": stat_rms,
        "nmb": stat_nmb,
        "mnmb": stat_mnmb,
        "mb": stat_mb,
        "mab": stat_mab,
        "fge": stat_fge,
        "R": stat_R,
        "R_spearman": stat_R_spearman,
        "R_kendall": stat_R_kendall,
    }


def calculate_statistics(
    data,
    ref_data,
    statistics: Mapping[str, StatisticsCalculator] | None = None,
    min_num_valid=1,
    weights: list | None = None,
    drop_stats=None,
    lowlim=None,
    highlim=None,
) -> StatsDict:
    """Calc statistical properties from two data arrays

    Filters data, calculates statistics and filters the resulting dict. If
    `statistics` is None the following statistics will be calculated by
    default.

        - Mean value of both arrays -> refdata_mean, data_mean
        - Standard deviation of both arrays -> refdata_std, data_std
        - RMS (Root mean square) -> rms
        - NMB (Normalised mean bias) -> nmb
        - MNMB (Modified normalised mean bias) -> mnmb
        - MB (Mean Bias) -> mb
        - MAB (Mean Absolute Bias) -> mab
        - FGE (Fractional gross error) -> fge
        - R (Pearson correlation coefficient) -> R
        - R_spearman (Spearman corr. coeff) -> R_spearman
        - R_kendall (Kendall's tau) ->

    Note
    ----
    Nans are removed from the input arrays, information about no. of removed
    points can be inferred from keys `totnum` and `num_valid` in return dict.

    Parameters
    ----------
    data : ndarray
        array containing data, that is supposed to be compared with reference
        data
    ref_data : ndarray
        array containing data, that is used to compare `data` array with
    statistics : dict[str, StatisticsCalculator]
        mapping between statistics name and Callable to calculate that statistics.
    lowlim : float
        lower end of considered value range (e.g. if set 0, then all datapoints
        where either ``data`` or ``ref_data`` is smaller than 0 are removed).
        Deprecated. Use data_filters with FilterByLimit instead.
    highlim : float
        upper end of considered value range
        Deprecated. Use data_filters with FilterByLimit instead.
    min_num_valid : int
        minimum number of valid measurements required to compute statistical
        parameters. Stat will be returned as NaN if length(data) is below this threshold.
    weights: ndarray
        array containing weights if computing weighted statistics.
    drop_stats: tuple
        tuple which drops the provided statistics from computed json files.
        For example, setting drop_stats = ("mb", "mab"), results in json files
        in hm/ts with entries which do not contain the mean bias and mean
        absolute bias, but the other statistics are preserved.
    Returns
    -------
    StatsDict
        dictionary containing computed statistics. Return mapping is the same as
        provided in `statistics`. Additionally a value for `totnum` (the unfiltered
        length of the data) and `weighted` (whether calculations was performed
        weighted or not) will be included.

    Raises
    ------
    ValueError
        if either of the input arrays has dimension other than 1
    ValueError
        if length of data is different to length of ref_data.
    ValueError
        if length of weights is different to length of data (assuming weights != None)
    """
    data = np.asarray(data)
    ref_data = np.asarray(ref_data)

    # Validation
    if not data.ndim == 1 or not ref_data.ndim == 1:
        raise ValueError("Invalid input. Data arrays must be one dimensional")

    if len(data) != len(ref_data):
        raise ValueError("Length mismatch between data and ref_data.")

    if weights is not None:
        if len(weights) != len(data):
            raise ValueError("Invalid input. Length of weights must match length of data.")

    # Set defaults
    if statistics is None:
        statistics = _get_default_statistic_config()

    data_filters = [FilterNaN()]
    if (lowlim is not None) or (highlim is not None):
        data_filters.append(FilterByLimit(lowlim=lowlim, highlim=highlim))

    stats_filters = None
    if drop_stats is not None:
        stats_filters = [FilterDropStats(drop_stats)]

    result = dict()

    result["totnum"] = len(data)
    result["weighted"] = False if weights is None else True

    data, ref_data, weights = _filter_data(data, ref_data, weights, data_filters)

    result["num_valid"] = float(len(data))

    if weights is None:
        weights = np.repeat([1], len(data))

    if len(weights) > 0:
        weights = weights / np.max(weights)

    additional_stats = _prepare_statistics(
        data, ref_data, weights, statistics, min_numvalid=min_num_valid
    )
    result.update(additional_stats)

    result = _filter_stats(result, stats_filters)

    return result


def _init_stats_dummy(drop_stats=None):
    # dummy for statistics dictionary for locations without data
    stats_dummy = {}
    for k in calculate_statistics([1], [1], drop_stats=drop_stats):
        stats_dummy[k] = np.nan

    # Test to make sure these variables are defined even when yearly and season != all
    stats_dummy["R_spatial_mean"] = np.nan
    stats_dummy["R_spatial_median"] = np.nan
    stats_dummy["R_temporal_mean"] = np.nan
    stats_dummy["R_temporal_median"] = np.nan

    return stats_dummy



from .output_paths import OutputPaths
from .model_maps_setup import ModelMapsSetup
from .statistics_setup import StatisticsSetup
from .time_setup import TimeSetup
from .cams2_83_setup import CAMS2_83Setup
from .web_display_setup import WebDisplaySetup
from .experiment_info import ExperimentInfo
from .eval_run_options import EvalRunOptions
from .eval_setup import EvalSetup
from .units_setup import UnitsSetup



import pytest

from pyaerocom._lowlevel_helpers import (
    ConstrainedContainer,
    NestedContainer,
    check_dir_access,
    invalid_input_err_str,
    sort_dict_by_name,
    str_underline,
)


@pytest.mark.parametrize("title", ["", "Bla", "Hello"])
@pytest.mark.parametrize("indent", [0, 4, 10])
def test_str_underline(title: str, indent: int):
    lines = str_underline(title, indent).split("\n")
    assert len(lines) == 2
    assert len(lines[0]) == len(lines[1]) == len(title) + indent
    assert lines[0].endswith(title)
    assert lines[1].endswith("-" * len(title))
    assert lines[0][:indent] == lines[1][:indent] == " " * indent


class Constrainer(ConstrainedContainer):
    def __init__(self):
        self.bla = 42
        self.blub = "str"
        self.opt = None


class NestedData(NestedContainer):
    def __init__(self):
        self.bla = dict(a=1, b=2)
        self.blub = dict(c=3, d=4)
        self.d = 42


def test_invalid_input_err_str():
    st = invalid_input_err_str("bla", "42", (42, 43))
    assert st == "Invalid input for bla (42), choose from (42, 43)"


@pytest.mark.parametrize("dir,val", [(".", True), ("/bla/blub", False), (42, False)])
def test_check_dir_access(dir, val):
    assert check_dir_access(dir) == val


def test_Constrainer():
    cont = Constrainer()
    assert cont.bla == 42
    assert cont.blub == "str"
    assert cont.opt is None


def test_NestedData():
    cont = NestedData()
    assert cont.bla == dict(a=1, b=2)
    assert cont.blub == dict(c=3, d=4)


@pytest.mark.parametrize("kwargs", [dict(), dict(bla=400), dict(bla=45, opt={})])
def test_ConstrainedContainer_update(kwargs):
    cont = Constrainer()
    cont.update(**kwargs)
    for key, val in kwargs.items():
        assert cont[key] == val


@pytest.mark.parametrize("kwargs", [dict(blaaaa=400)])
def test_ConstrainedContainer_update_error(kwargs):
    cont = Constrainer()
    with pytest.raises(ValueError):
        cont.update(**kwargs)


def test_NestedData_keys_unnested():
    cont = NestedData()
    keys = cont.keys_unnested()
    assert sorted(keys) == ["a", "b", "bla", "blub", "c", "d", "d"]


def test_NestedData___getitem__():
    cont = NestedData()
    assert cont["d"] == 42


def test_NestedData___getitem___error():
    cont = NestedData()
    with pytest.raises(KeyError):
        cont["a"]


@pytest.mark.parametrize("kwargs", [dict(bla=42), dict(a=400), dict(d=400)])
def test_NestedData_update(kwargs):
    cont = NestedData()
    cont.update(**kwargs)
    for key, value in kwargs.items():
        if key in cont.__dict__:  # toplevel entry
            assert cont[key] == value
        for val in cont.values():
            if isinstance(val, dict) and key in val:
                assert val[key] == value


def test_NestedData_update_error():
    cont = NestedData()
    with pytest.raises(AttributeError) as e:
        cont.update(abc=400)
    assert str(e.value) == "invalid key abc"


@pytest.mark.parametrize("input", [{"b": 1, "a": 2, "kl": 42}])
@pytest.mark.parametrize(
    "pref_list,output_keys",
    [
        ([], ["a", "b", "kl"]),
        (["blaaa"], ["a", "b", "kl"]),
        (["kl"], ["kl", "a", "b"]),
        (["kl", "b"], ["kl", "b", "a"]),
    ],
)
def test_sort_dict_by_name(input, pref_list, output_keys):
    sorted = sort_dict_by_name(input, pref_list)
    assert list(sorted.keys()) == output_keys



import numpy as np
import pytest

from pyaerocom.stats.data_filters import FilterByLimit, FilterNaN


@pytest.mark.parametrize(
    "data,ref_data,expected_mask",
    (
        pytest.param(
            np.asarray([1, 2, 3, 4]), np.asarray([1, 2, 3, 4]), np.repeat([True], 4), id="no-nan"
        ),
        pytest.param(
            np.repeat([np.nan], 10), np.repeat([np.nan], 10), np.repeat(False, 10), id="only-nan"
        ),
        pytest.param(
            np.asarray([np.nan, 5] * 6),
            np.asarray([np.nan, 5, 3] * 4),
            np.asarray([False, True, False, False, False, True] * 2),
            id="mixed-nan",
        ),
    ),
)
def test_filter_nan(data, ref_data, expected_mask):
    filternan = FilterNaN()

    mask = filternan(data, ref_data, None)

    assert len(mask) == len(data)
    assert np.array_equal(mask, expected_mask, equal_nan=True)


@pytest.mark.parametrize(
    "data,ref_data,lowlim,highlim,expected_mask",
    (
        pytest.param(
            np.asarray([1, 2, 3, 4] * 5),
            np.asarray([5, 4] * 10),
            None,
            None,
            np.asarray([True] * 20),
            id="no-limit",
        ),
        pytest.param(
            np.asarray([1, 2, 3, 4] * 5),
            np.asarray([1, 2, 3, 4] * 5),
            3,
            None,
            np.asarray([False, False, False, True] * 5),
            id="low-limit",
        ),
        pytest.param(
            np.asarray([1, 2, 3, 4] * 5),
            np.asarray([1, 2, 3, 4] * 5),
            None,
            2,
            np.asarray([True, False, False, False] * 5),
            id="high-limit",
        ),
        pytest.param(
            np.asarray([1, 2, 3, 4] * 5),
            np.asarray([1, 2, 3, 4] * 5),
            1,
            4,
            np.asarray([False, True, True, False] * 5),
            id="both-limits",
        ),
    ),
)
def test_filterbylimit(data, ref_data, lowlim, highlim, expected_mask):
    filterbylimit = FilterByLimit(lowlim=lowlim, highlim=highlim)

    mask = filterbylimit(data, ref_data, None)

    assert len(mask) == len(data)
    assert np.array_equal(mask, expected_mask, equal_nan=True)



# Mypy errors because it doesn't understand the total_ordering
# decorator used for TsType. This suppresses that error:
# mypy: disable-error-code=operator

import calendar
from collections import Counter

from cf_units import Unit
import iris
from pyaerocom._warnings import ignore_warnings
from pyaerocom.exceptions import TemporalResolutionError
from . import TsType
from ..constants import SECONDS_IN_DAY
from .time_config import (
    GREGORIAN_BASE,
    TS_TYPE_SECS,
    microsec_units,
    millisec_units,
    sec_units,
    min_units,
    hr_units,
    day_units,
    TS_TYPE_DATETIME_CONV,
)
import pandas as pd
import numpy as np
from datetime import datetime, date
from datetime import MINYEAR

import logging


logger = logging.getLogger(__name__)


def seconds_in_periods(timestamps, ts_type):
    """
    Calculates the number of seconds for each period in timestamps.

    Parameters
    ----------
    timestamps : numpy.datetime64 or numpy.ndarray
        Either a single datetime or an array of datetimes.
    ts_type : str
        Frequency of timestamps.

    Returns
    -------
    np.array :
        Array with same length as timestamps containing number of seconds for
        each period.
    """

    ts_type = TsType(ts_type)
    if isinstance(timestamps, np.datetime64):
        timestamps = np.array([timestamps])
    if isinstance(timestamps, np.ndarray):
        timestamps = [to_pandas_timestamp(timestamp) for timestamp in timestamps]
    # From here on timestamps should be a numpy array containing pandas Timestamps
    if ts_type >= TsType("monthly"):
        if ts_type == TsType("monthly"):
            days_in_months = np.array([timestamp.days_in_month for timestamp in timestamps])
            seconds = days_in_months * SECONDS_IN_DAY
            return seconds
        if ts_type == TsType("daily"):
            return SECONDS_IN_DAY * np.ones_like(timestamps)
        raise NotImplementedError("Only yearly, monthly and daily frequencies implemented.")

    if ts_type == TsType("yearly"):
        days_in_year = [365 + calendar.isleap(ts.year) for ts in timestamps]
        seconds = np.array(days_in_year) * SECONDS_IN_DAY
        return seconds

    raise TemporalResolutionError(f"Unknown TsType: {ts_type}")


def is_year(val) -> bool:
    """Check if input is / may be year

    Parameters
    ----------
    val
        input that is supposed to be checked

    Returns
    -------
    bool
        True if input is a number between -2000 and 10000, else False
    """
    try:
        if -2000 < int(val) < 10000:
            return True
    except ValueError:
        pass

    return False


def get_tot_number_of_seconds(ts_type: str, dtime: pd.Series | None = None):
    """Get total no. of seconds for a given frequency

    ToDo
    ----
    This method needs revision and can be solved simpler probably

    Parameters
    ----------
    ts_type : str or TsType
        frequency for which number of seconds is supposed to be retrieved
    dtime : TYPE, optional
        DESCRIPTION. The default is None.

    Raises
    ------
    AttributeError
        DESCRIPTION.

    Returns
    -------
    TYPE
        DESCRIPTION.

    """

    ts_tpe = TsType(ts_type)

    if ts_tpe >= TsType("monthly"):
        if dtime is None:
            raise AttributeError(
                "For frequencies larger than or eq. monthly you"
                + " need to provide dtime in order to compute the number of second."
            )
        if not ts_type == "monthly":
            raise NotImplementedError("Can only handle monthly so far...")

        # find seconds from dtime
        # TODO generalize this
        days_in_month = dtime.dt.daysinmonth

        return days_in_month * SECONDS_IN_DAY

    return TS_TYPE_SECS[ts_type]


def to_datetime64(value):
    """Convert input value to numpy.datetime64

    Parameters
    ----------
    value
        input value that is supposed to be converted, needs to be either str,
        datetime.datetime, pandas.Timestamp or an integer specifying the
        desired year.

    Returns
    -------
    datetime64
        input timestamp converted to datetime64
    """
    if isinstance(value, np.datetime64):
        return value

    try:
        return to_pandas_timestamp(value).to_datetime64()
    except Exception as e:
        raise ValueError(f"Failed to convert {value} to datetime64 objectError: {repr(e)}")


def cftime_to_datetime64(times, cfunit=None, calendar=None):
    """Convert numerical timestamps with epoch to numpy datetime64

    This method was designed to enhance the performance of datetime conversions
    and is based on the corresponding information provided in the cftime
    package (`see here <https://github.com/Unidata/cftime/blob/master/cftime/
    _cftime.pyx>`__). Particularly, this object does, what the :func:`num2date`
    therein does, but faster, in case the time stamps are not defined on a non
    standard calendar.

    Parameters
    ----------
    times : :obj:`list` or :obj:`ndarray` or :obj:`iris.coords.DimCoord`
        array containing numerical time stamps (relative to basedate of
        ``cfunit``). Can also be a single number.
    cfunit : :obj:`str` or :obj:`Unit`, optional
        CF unit string (e.g. day since 2018-01-01 00:00:00.00000000 UTC) or
        unit. Required if `times` is not an instance of
        :class:`iris.coords.DimCoord`
    calendar : :obj:`str`, optional
        string specifying calendar (only required if ``cfunit`` is of type
        ``str``).

    Returns
    -------
    ndarray
        numpy array containing timestamps as datetime64 objects

    Raises
    ------
    ValueError
        if cfunit is ``str`` and calendar is not provided or invalid, or if
        the cfunit string is invalid

    Example
    -------

    >>> cfunit_str = 'day since 2018-01-01 00:00:00.00000000 UTC'
    >>> cftime_to_datetime64(10, cfunit_str, "gregorian")
    array(['2018-01-11T00:00:00.000000'], dtype='datetime64[us]')
    """
    if isinstance(times, iris.coords.DimCoord):  # special case
        times, cfunit = times.points, times.units
    try:
        len(times)
    except Exception:
        times = [times]
    if isinstance(cfunit, str):
        if calendar is None:
            raise ValueError(
                "Require specification of calendar for conversion into datetime64 objects"
            )
        cfunit = Unit(cfunit, calendar)  # raises Error if calendar is invalid
    if not isinstance(cfunit, Unit):
        raise ValueError(
            "Please provide cfunit either as instance of class cf_units.Unit or as a string"
        )
    calendar = cfunit.calendar
    basedate = cfunit.num2date(0)
    if (calendar == "proleptic_gregorian" and basedate.year >= MINYEAR) or (
        calendar in ["gregorian", "standard"] and basedate > GREGORIAN_BASE
    ):
        cfu_str = cfunit.origin

        res = cfu_str.split()[0].lower()
        if res in microsec_units:
            tstr = "us"
        elif res in millisec_units:
            tstr = "ms"
        elif res in sec_units:
            tstr = "s"
        elif res in min_units:
            tstr = "m"
        elif res in hr_units:
            tstr = "h"
        elif res in day_units:
            tstr = "D"
        else:
            raise ValueError("unsupported time units")

        basedate = np.datetime64(basedate)
        dt = np.asarray(np.asarray(times), dtype=f"timedelta64[{tstr}]")
        return basedate + dt
    else:
        return np.asarray([np.datetime64(t) for t in cfunit.num2date(times)])


@ignore_warnings(UserWarning, r"Parsing .* in DD/MM/YYYY format")
def to_pandas_timestamp(value):
    """Convert input to instance of :class:`pandas.Timestamp`

    Parameters
    ----------
    value
        input value that is supposed to be converted to time stamp

    Returns
    --------
    pandas.Timestamp
    """
    if isinstance(value, np.str_):
        value = str(value)
    if isinstance(value, pd.Timestamp):
        return value
    if isinstance(value, str | np.datetime64 | datetime | date):
        return pd.Timestamp(value)

    try:
        numval = int(value)
        if not 0 <= numval <= 10000:
            raise ValueError("Could not infer valid year from numerical time input")
        return pd.Timestamp(str(numval))
    except Exception as e:
        raise ValueError(f"Failed to convert {value} to Timestamp: {repr(e)}")


def infer_time_resolution(time_stamps, dt_tol_percent=5, minfrac_most_common=0.8) -> TsType:
    """Infer time resolution based on input time-stamps

    Calculates time difference *dt* between consecutive timestamps provided via
    input array or list. Then it counts the most common *dt* (e.g. 86400 s for
    daily). Before inferring the frequency it then checks all other *dts*
    occurring in the input array to see if they are within a certain interval
    around the most common one (e.g. +/- 5% as default, via arg
    `dt_tol_percent`), that is, 86390 would be included if most common dt is
    86400 s but not 80000s. Then it checks if the number of *dts* that
    are within that tolerance level around the most common *dt* exceed a
    certain fraction (arg `minfrac_most_common`) of the total number of *dts*
    that occur in the input array (default is 80%). If that is the case, the
    most common frequency is attempted to be derived using
    :func:`TsType.from_total_seconds` based on the most common *dt* (in this
    example that would be *daily*).


    Parameters
    ----------
    time_stamps : pandas.DatetimeIndex, or similar
        list of time stamps
    dt_tol_percent : int
        tolerance in percent of accepted range of time diffs with respect to
        most common time difference.
    minfrac_most_common : float
        minimum required fraction of time diffs that have to be equal to, or
        within tolerance range, the most common time difference.


    Raises
    ------
    TemporalResolutionError
        if frequency cannot be derived.

    Returns
    -------
    str
        inferred frequency
    """
    if not isinstance(time_stamps, pd.DatetimeIndex):
        time_stamps = pd.DatetimeIndex(time_stamps)
    vals = time_stamps.values

    dts = (vals[1:] - vals[:-1]).astype("timedelta64[s]").astype(int)

    if np.min(dts) < 0:
        raise TemporalResolutionError("Nasa Ames file contains neg. meas periods...")

    counts = Counter(dts).most_common()
    most_common_dt, most_common_num = counts[0]
    num_within_tol = most_common_num
    lower = most_common_dt * (100 - dt_tol_percent) / 100
    upper = most_common_dt * (100 + dt_tol_percent) / 100
    for dt, num in counts[1:]:
        if lower <= dt <= upper:
            num_within_tol += num
    frac_ok = num_within_tol / len(dts)
    if not frac_ok > minfrac_most_common:
        raise TemporalResolutionError("Failed to infer ts_type")
    tst = TsType.from_total_seconds(most_common_dt)
    return tst


def test_cftime_to_datetime64():
    pass


def datetime2str(time, ts_type=None):
    conv = TS_TYPE_DATETIME_CONV[ts_type]
    if is_year(time):
        return str(time)
    try:
        time = to_pandas_timestamp(time).strftime(conv)
    except pd.errors.OutOfBoundsDatetime:
        logger.warning(f"Failed to convert time {time} to string")
    return time



import logging

from pyaerocom import ColocatedData
from pyaerocom.aeroval._processing_base import HasColocator, ProcessingEngine
from pyaerocom.aeroval.obsentry import ObsEntry
from pyaerocom.aeroval.modelentry import ModelEntry
from pyaerocom.aeroval.coldatatojson_engine import ColdataToJsonEngine


from pyaerocom.colocation.colocator import Colocator
from pyaerocom.colocation.colocation_setup import ColocationSetup


logger = logging.getLogger(__name__)


class BulkFractionEngine(ProcessingEngine, HasColocator):
    def run(self, var_list: list[str] | str | None, model_name: str, obs_name: str):
        self.sobs_cfg = self.cfg.obs_cfg.get_entry(obs_name)
        self.smodel_cfg = self.cfg.model_cfg.get_entry(model_name)

        if var_list is None:
            var_list = self.sobs_cfg.obs_vars
        elif isinstance(var_list, str):
            var_list = [var_list]
        elif not isinstance(var_list, list):
            raise ValueError(f"invalid input for var_list: {var_list}.")

        files_to_convert = []
        for var_name in var_list:
            bulk_vars = self._get_bulk_vars(var_name, self.sobs_cfg)

            freq = self.sobs_cfg.ts_type
            cd, fp = self._run_var(
                model_name, obs_name, var_name, bulk_vars, freq, self.sobs_cfg, self.smodel_cfg
            )

            files_to_convert.append(fp)

        engine = ColdataToJsonEngine(self.cfg)
        engine.run(files_to_convert)

    def _get_bulk_vars(self, var_name: str, cfg: ObsEntry) -> list:
        bulk_vars = cfg.bulk_options
        if var_name not in bulk_vars:
            raise KeyError(f"Could not find bulk vars entry for {var_name}")

        if len(bulk_vars[var_name].vars) != 2:
            raise ValueError(
                f"(Only) 2 entries must be present for bulk vars to calculate fraction for {var_name}. Found {bulk_vars[var_name]}"
            )

        return bulk_vars[var_name].vars

    def _run_var(
        self,
        model_name: str,
        obs_name: str,
        var_name: str,
        bulk_vars: list,
        freq: str,
        obs_entry: ObsEntry,
        model_entry: ModelEntry,
    ) -> tuple[ColocatedData, str]:
        model_exists = obs_entry.bulk_options[var_name].model_exists

        cols = self.get_colocators(bulk_vars, var_name, freq, model_name, obs_name, model_exists)

        coldatas = []
        for col in cols:
            if len(list(col.keys())) != 1:
                raise ValueError(
                    "Found more than one colocated object when trying to run bulk variable"
                )
            bv = list(col.keys())[0]
            coldatas.append(col[bv].run(bv))

        num_name, denum_name = bulk_vars[0], bulk_vars[1]
        num_col = cols[0][num_name].run(num_name)
        denum_col = cols[1][denum_name].run(denum_name)

        model_num_name, model_denum_name = self._get_model_var_names(
            var_name, bulk_vars, model_exists, model_entry
        )

        cd = self._combine_coldatas(
            num_col[model_num_name][num_name],
            denum_col[model_denum_name][denum_name],
            var_name,
            obs_entry,
        )

        num_colocator = cols[0][num_name]

        fp = cd.to_netcdf(
            out_dir=num_colocator.output_dir,
            savename=cd._aerocom_savename(
                var_name,
                obs_name,
                var_name,
                model_name,
                num_colocator.get_start_str(),
                num_colocator.get_stop_str(),
                freq,
                num_colocator.colocation_setup.filter_name,
                None,  # cd.data.attrs["vert_code"],
            ),
        )
        return cd, fp

    def _combine_coldatas(
        self,
        num_coldata: ColocatedData,
        denum_coldata: ColocatedData,
        var_name: str,
        obs_entry: ObsEntry,
    ) -> ColocatedData:
        mode = obs_entry.bulk_options[var_name].mode
        model_exists = obs_entry.bulk_options[var_name].model_exists
        units = obs_entry.bulk_options[var_name].units

        if mode == "fraction":
            new_data = num_coldata.data / denum_coldata.data

        elif mode == "product":
            new_data = num_coldata.data * denum_coldata.data
        else:
            raise ValueError(f"Mode must be either fraction of product, and not {mode}")
        if model_exists:
            # TODO: Unsure if this works!!!
            new_data[1] = num_coldata.data[1].where(new_data[1])

        cd = ColocatedData(new_data)

        cd.data.attrs = num_coldata.data.attrs
        cd.data.attrs["var_name"] = [var_name, var_name]
        cd.data.attrs["var_units"] = [units, units]
        cd.metadata["var_name_input"] = [var_name, var_name]
        return cd

    def _get_model_var_names(
        self, var_name: str, bulk_vars: list[str], model_exists: bool, model_entry: ModelEntry
    ) -> tuple[str]:
        num_name, denum_name = bulk_vars[0], bulk_vars[1]
        if model_exists:
            num_name, denum_name = var_name, var_name

        model_use_vars = model_entry.model_use_vars
        if model_use_vars != {}:
            num_name, denum_name = model_use_vars[num_name], model_use_vars[denum_name]

        return num_name, denum_name

    def get_colocators(
        self,
        bulk_vars: list,
        var_name: str,
        freq: str,
        model_name: str = None,
        obs_name: str = None,
        model_exists: bool = False,
    ) -> list[dict[str | Colocator]]:
        """
        Instantiate colocation engine

        Parameters
        ----------
        model_name : str, optional
            name of model. The default is None.
        obs_name : str, optional
            name of obs. The default is None.

        Returns
        -------
        Colocator

        """
        col_cfg = {**self.cfg.colocation_opts.model_dump()}
        outdir = self.cfg.path_manager.get_coldata_dir()
        col_cfg["basedir_coldata"] = outdir

        if model_name:
            mod_cfg = self.cfg.get_model_entry(model_name)
            col_cfg["model_cfg"] = mod_cfg

            # Hack and at what lowlevel_helpers's import_from was doing
            for key, val in mod_cfg.items():
                if key in ColocationSetup.model_fields:
                    col_cfg[key] = val
        if obs_name:
            obs_cfg = self.cfg.get_obs_entry(obs_name)
            pyaro_config = obs_cfg["obs_config"] if "obs_config" in obs_cfg else None
            col_cfg["obs_config"] = pyaro_config

            # Hack and at what lowlevel_helpers's import_from was doing
            for key, val in obs_cfg.model_dump().items():
                if key in ColocationSetup.model_fields:
                    col_cfg[key] = val

            col_cfg["add_meta"].update(diurnal_only=self.cfg.get_obs_entry(obs_name).diurnal_only)
        cols = []
        col_cfg["ts_type"] = freq
        for bulk_var in bulk_vars:
            col_cfg["obs_vars"] = bulk_var
            if model_exists:
                col_cfg["model_use_vars"] = {
                    bulk_var: var_name,
                }
            col_stp = ColocationSetup(**col_cfg)
            col = Colocator(col_stp)

            cols.append({bulk_var: col})

        return cols



"""
Methods and / or classes to perform colocation
"""

import logging
import os

import numpy as np
import pandas as pd
import xarray as xr
from geonum.atmosphere import pressure

from pyaerocom import __version__ as pya_ver
from pyaerocom import const
from pyaerocom._lowlevel_helpers import RegridResDeg
from pyaerocom.climatology_config import ClimatologyConfig
from pyaerocom.exceptions import (
    DimensionOrderError,
    MetaDataError,
    TemporalResolutionError,
    TimeMatchError,
    VariableDefinitionError,
    VarNotAvailableError,
)
from pyaerocom.filter import Filter
from pyaerocom.griddeddata import GriddedData
from pyaerocom.ungridded_data_container import UngriddedDataContainer
from pyaerocom.units.datetime import get_lowest_resolution, to_pandas_timestamp
from pyaerocom.helpers import (
    isnumeric,
    make_datetime_index,
)
from pyaerocom.time_resampler import TimeResampler
from pyaerocom.units.datetime import TsType

from .colocated_data import ColocatedData

logger = logging.getLogger(__name__)


def resolve_var_name(data: GriddedData) -> tuple[str, str]:
    """
    Check variable name of `GriddedData` against AeroCom default

    Checks whether the variable name set in the data corresponds to the
    AeroCom variable name, or whether it is an alias. Returns both the
    variable name set and the AeroCom variable name.

    Parameters
    ----------
    data : GriddedData
        Data to be checked.

    Returns
    -------
    str
        variable name as set in data (may be alias, but may also be AeroCom
        variable name, in which case first and second return parameter are the
        same).
    str
        corresponding AeroCom variable name

    """

    var = data.var_name
    try:
        vardef = const.VARS[var]
    except VariableDefinitionError:
        vardef = data.register_var_glob()

    return (var, vardef.var_name_aerocom)


def _regrid_gridded(gridded, regrid_scheme: str, regrid_res_deg: RegridResDeg):
    """
    Regrid instance of `GriddedData`

    Makes sure to handle different input options for `regrid_res_deg`.

    Parameters
    ----------
    gridded : GriddedData
        instance of :class:`GriddedData` that is supposed to be regridded.
    regrid_scheme : str
        iris scheme used for regridding (defaults to area weighted regridding)
    regrid_res_deg : int or dict, optional
        regrid resolution in degrees. If specified, the input gridded data
        objects will be regridded in lon / lat dimension to the input
        resolution (if input is integer, both lat and lon are regridded to that
        resolution, if input is dict, use keys `lat_res_deg` and `lon_res_deg`
        to specify regrid resolutions, respectively).

    Raises
    ------
    ValueError
        If input for `regrid_res_deg` is invalid.

    Returns
    -------
    GriddedData
        regridded data object

    """
    if not isinstance(regrid_res_deg, dict):  # at runtime RegridResDeg is a dict
        if not isnumeric(regrid_res_deg):
            raise ValueError(
                "Invalid input for regrid_res_deg. Need integer or dict specifying lat and lon res"
            )
        regrid_res_deg = dict(lat_res_deg=regrid_res_deg, lon_res_deg=regrid_res_deg)

    return gridded.regrid(scheme=regrid_scheme, **regrid_res_deg)


def _ensure_gridded_gridded_same_freq(data, data_ref, min_num_obs, resample_how):
    """
    Make sure 2 input gridded data objects are in the same frequency

    Checks if both input data objects are in the same frequency, and if not,
    downsample the one with higher frequency accordingly.

    Parameters
    ----------
    data : GriddedData
        first data object.
    data_ref : GriddedData
        second data object.
    min_num_obs : int or dict, optional
        Minimum number of observations for resampling.
    resample_how : str or dict, optional
        Resampling aggregators used.

    Returns
    -------
    GriddedData
        first data object.
    GriddedData
        second data object.
    str
        sampling frequency of both data objects.

    """
    ts_type_data = data.ts_type
    ts_type_data_ref = data_ref.ts_type
    if ts_type_data != ts_type_data_ref:
        # ref data is in higher resolution
        if TsType(ts_type_data_ref) > TsType(ts_type_data):
            data_ref = data_ref.resample_time(
                ts_type_data, min_num_obs=min_num_obs, how=resample_how
            )
        else:
            data = data.resample_time(ts_type_data_ref, min_num_obs=min_num_obs, how=resample_how)
    return data, data_ref, data.ts_type


def colocate_gridded_gridded(
    data,
    data_ref,
    ts_type=None,
    start=None,
    stop=None,
    filter_name=None,
    regrid_res_deg: float | RegridResDeg | None = None,
    regrid_scheme: str = "areaweighted",
    update_baseyear_gridded=None,
    min_num_obs=None,
    colocate_time=False,
    resample_how=None,
    **kwargs,
):
    """Colocate 2 gridded data objects

    TODO
    ----
    - think about vertical dimension (vert_scheme input not used at the moment)

    Parameters
    ----------
    data : GriddedData
        gridded data (e.g. model results)
    data_ref : GriddedData
        reference data (e.g. gridded satellite object) that is co-located with
        `data`.
        observation data or other model)
    ts_type : str, optional
        desired temporal resolution of output colocated data (e.g. "monthly").
        Defaults to None, in which case the highest possible resolution is
        used.
    start : str or datetime64 or similar, optional
        start time for colocation, if None, the start time of the input
        :class:`GriddedData` object is used
    stop : str or datetime64 or similar, optional
        stop time for colocation, if None, the stop time of the input
        :class:`GriddedData` object is used
    filter_name : str, optional
        string specifying filter used (cf. :class:`pyaerocom.filter.Filter` for
        details). If None, then it is set to 'ALL-wMOUNTAINS', which
        corresponds to no filtering (world with mountains).
        Use ALL-noMOUNTAINS to exclude mountain sites.
    regrid_res_deg : int or dict, optional
        regrid resolution in degrees. If specified, the input gridded data
        objects will be regridded in lon / lat dimension to the input
        resolution (if input is integer, both lat and lon are regridded to that
        resolution, if input is dict, use keys `lat_res_deg` and `lon_res_deg`
        to specify regrid resolutions, respectively).
    regrid_scheme : str
        iris scheme used for regridding (defaults to area weighted regridding)
    update_baseyear_gridded : int, optional
        optional input that can be set in order to redefine the time dimension
        in the first gridded data object `data`to be analysed. E.g., if the
        data object is a climatology (one year of data) that has set the base
        year of the time dimension to a value other than the specified input
        start / stop time this may be used to update the time in order to make
        co-location possible.
    min_num_obs : int or dict, optional
        minimum number of observations for resampling of time
    colocate_time : bool
        if True and if original time resolution of data is higher than desired
        time resolution (`ts_type`), then both datasets are colocated in time
        *before* resampling to lower resolution. NOTE: If there are missing data
        in the model, having this true will interpolate the missing model data!
    resample_how : str or dict
        string specifying how data should be aggregated when resampling in time.
        Default is "mean". Can also be a nested dictionary, e.g.
        resample_how={'daily': {'hourly' : 'max'}} would use the maximum value
        to aggregate from hourly to daily, rather than the mean.
    **kwargs
        additional keyword args (not used here, but included such that factory
        class can handle different methods with different inputs)

    Returns
    -------
    ColocatedData
        instance of colocated data

    """
    if filter_name is None:
        filter_name = const.DEFAULT_REG_FILTER

    if update_baseyear_gridded is not None:
        # update time dimension in gridded data
        data.base_year = update_baseyear_gridded

    if regrid_res_deg is not None:
        data_ref = _regrid_gridded(data_ref, regrid_scheme, regrid_res_deg)
    # perform regridding
    if data.lon_res < data_ref.lon_res:  # obs has lower resolution
        data = data.regrid(data_ref, scheme=regrid_scheme)
    else:
        data_ref = data_ref.regrid(data, scheme=regrid_scheme)

    ts_type_src = [data_ref.ts_type, data.ts_type]
    # time resolution of dataset to be analysed
    data, data_ref, data_ts_type = _ensure_gridded_gridded_same_freq(
        data, data_ref, min_num_obs, resample_how
    )
    # now both are in same temporal resolution

    # input ts_type is not specified or model is in lower resolution
    # than input ts_type -> use model frequency to colocate
    if ts_type is None or TsType(data_ts_type) < TsType(ts_type):
        ts_type = data_ts_type

    # 1. match model data with potential input start / stop and update if
    # applicable
    start, stop = check_time_ival(data, start, stop)
    # 2. narrow it down with obsdata availability, if applicable
    start, stop = check_time_ival(data_ref, start, stop)

    data = data.crop(time_range=(start, stop))
    data_ref = data_ref.crop(time_range=(start, stop))

    # perform region extraction (if applicable)
    regfilter = Filter(name=filter_name)
    data = regfilter(data)
    data_ref = regfilter(data_ref)

    files_ref = [os.path.basename(x) for x in data_ref.from_files]
    files = [os.path.basename(x) for x in data.from_files]

    var, var_aerocom = resolve_var_name(data)
    var_ref, var_ref_aerocom = resolve_var_name(data_ref)
    meta = {
        "data_source": [data_ref.data_id, data.data_id],
        "var_name": [var_ref_aerocom, var_aerocom],
        "var_name_input": [var_ref, var],
        "ts_type": data_ts_type,
        "filter_name": filter_name,
        "ts_type_src": ts_type_src,
        "var_units": [str(data_ref.units), str(data.units)],
        "data_level": 3,
        "revision_ref": data_ref.data_revision,
        "from_files": files,
        "from_files_ref": files_ref,
        "colocate_time": colocate_time,
        "obs_is_clim": False,
        "pyaerocom": pya_ver,
        "min_num_obs": min_num_obs,
        "resample_how": resample_how,
    }

    data_np = data.grid.data
    if isinstance(data_np, np.ma.core.MaskedArray):
        data_np = data_np.filled(np.nan)
    data_ref_np = data_ref.grid.data
    if isinstance(data_ref_np, np.ma.core.MaskedArray):
        data_ref_np = data_ref_np.filled(np.nan)
    arr = np.asarray((data_ref_np, data_np))
    time = data.time_stamps().astype("datetime64[ns]")
    lats = data.latitude.points
    lons = data.longitude.points

    # create coordinates of DataArray
    coords = {
        "data_source": meta["data_source"],
        "time": time,
        "latitude": lats,
        "longitude": lons,
    }

    dims = ["data_source", "time", "latitude", "longitude"]

    coldata = ColocatedData(data=arr, coords=coords, dims=dims, name=data.var_name, attrs=meta)

    # add correct units for lat / lon dimensions
    coldata.latitude.attrs["standard_name"] = data.latitude.standard_name
    coldata.latitude.attrs["units"] = str(data.latitude.units)

    coldata.longitude.attrs["standard_name"] = data.longitude.standard_name
    coldata.longitude.attrs["units"] = str(data.longitude.units)

    if data_ts_type != ts_type:
        coldata = coldata.resample_time(
            to_ts_type=ts_type,
            colocate_time=colocate_time,
            min_num_obs=min_num_obs,
            how=resample_how,
            **kwargs,
        )
    return coldata


def check_time_ival(data, start, stop):
    # get start / stop of gridded data as pandas.Timestamp
    data_start = to_pandas_timestamp(data.start)
    data_stop = to_pandas_timestamp(data.stop)

    if start is None:
        start = data_start
    else:
        start = to_pandas_timestamp(start)
    if stop is None:
        stop = data_stop
    else:
        stop = to_pandas_timestamp(stop)

    if start < data_start:
        start = data_start
    if stop > data_stop:
        stop = data_stop
    # check overlap
    if stop < data_start or start > data_stop:
        raise TimeMatchError(
            f"Input time range {start}-{stop} does not overlap with data "
            f"range: {data_start}-{data_stop}"
        )
    return start, stop


def check_ts_type(data, ts_type):
    ts_type_data = TsType(data.ts_type)
    if ts_type is None:
        ts_type = ts_type_data
    elif isinstance(ts_type, str):
        ts_type = TsType(ts_type)
    if ts_type > ts_type_data:
        # desired output frequency ts_type is higher resolution than frequency
        # of data (e.g. desired output is hourly but data is daily, update
        # output ts_type)
        ts_type = ts_type_data
    return ts_type, ts_type_data


def _colocate_site_data_helper(
    stat_data,
    stat_data_ref,
    var,
    var_ref,
    ts_type,
    resample_how,
    min_num_obs,
    use_climatology_ref,
):
    """
    Helper method that colocates two timeseries from 2 StationData objects

    Used in main loop of :func:`colocate_gridded_ungridded`

    Parameters
    ----------
    stat_data : StationData
        first data object (usually the one that is to be compared with obs)
    stat_data_ref : StationData
        second data object (usually obs)
    var : str
        variable to be used from `stat_data`
    var_ref : str
        variable to be used from `stat_data_ref`
    ts_type : str
        output frequency
    resample_how : str or dict
        string specifying how data should be aggregated when resampling in time.
        Default is "mean". Can also be a nested dictionary, e.g.
        resample_how={'daily': {'hourly' : 'max'}} would use the maximum value
        to aggregate from hourly to daily, rather than the mean.
    min_num_obs : int or dict, optional
        minimum number of observations for resampling of time
    use_climatology_ref : ClimateConfig | bool, optional
        If provided, the climatology will be calculated from the config


    Raises
    ------
    TemporalResolutionError
        if obs sampling frequency is lower than desired output frequency

    Returns
    -------
    pandas.DataFrame
        dataframe containing the colocated input data (column names are
        data and ref)
    """

    # get grid and obs timeseries data (that may be sampled in arbitrary
    # time resolution, particularly the obs data)
    grid_ts = stat_data.resample_time(
        var, ts_type=ts_type, how=resample_how, min_num_obs=min_num_obs, inplace=True
    )[var]

    if isinstance(use_climatology_ref, ClimatologyConfig):
        obs_ts = stat_data_ref.calc_climatology(
            var_ref,
            start=use_climatology_ref.start,
            stop=use_climatology_ref.stop,
            min_num_obs=min_num_obs,
            clim_mincount=use_climatology_ref.min_count,
            resample_how=use_climatology_ref.resample_how,
            clim_freq=use_climatology_ref.freq,
            set_year=use_climatology_ref.set_year,
        )[var_ref]
    else:
        obs_ts = stat_data_ref.resample_time(
            var_ref,
            ts_type=ts_type,
            how=resample_how,
            min_num_obs=min_num_obs,
            inplace=True,
        )[var_ref]

    # fill up missing time stamps
    return pd.concat([obs_ts, grid_ts], axis=1, keys=["ref", "data"])


def _colocate_site_data_helper_timecol(
    stat_data,
    stat_data_ref,
    var,
    var_ref,
    ts_type,
    resample_how,
    min_num_obs,
    use_climatology_ref,
):
    """
    Helper method that colocates two timeseries from 2 StationData objects

    Other than :func:`_colocate_site_data_helper` this method applies time
    colocation in highest possible resolution (used if option `colocate_time`
    is active in colocation routine :func:`colocate_gridded_ungridded`).

    Used in main loop of :func:`colocate_gridded_ungridded`

    Parameters
    ----------
    stat_data : StationData
        first data object (usually the one that is to be compared with obs)
    stat_data_ref : StationData
        second data object (usually obs)
    var : str
        variable to be used from `stat_data`
    var_ref : str
        variable to be used from `stat_data_ref`
    ts_type : str
        output frequency
    resample_how : str or dict
        string specifying how data should be aggregated when resampling in time.
        Default is "mean". Can also be a nested dictionary, e.g.
        resample_how={'daily': {'hourly' : 'max'}} would use the maximum value
        to aggregate from hourly to daily, rather than the mean.
    min_num_obs : int or dict, optional
        minimum number of observations for resampling of time
    use_climatology_ref: ClimateConfig | bool
        if provided, NotImplementedError is raised

    Raises
    ------
    TemporalResolutionError
        if model or obs sampling frequency is lower than desired output frequency
    NotImplementedError
        if input arg `use_climatology_ref` is provided.

    Returns
    -------
    pandas.DataFrame
        dataframe containing the colocated input data (column names are
        data and ref)
    """
    if isinstance(use_climatology_ref, ClimatologyConfig):
        raise NotImplementedError(
            "Using observation climatology in colocation with option colocate_time is not available yet"
        )

    grid_tst = stat_data.get_var_ts_type(var)
    obs_tst = stat_data_ref.get_var_ts_type(var_ref)
    coltst = TsType(get_lowest_resolution(grid_tst, obs_tst))
    # =============================================================================
    #     if coltst.mulfac != 1:
    #         coltst = coltst.next_lower
    # =============================================================================

    stat_data.resample_time(
        var_name=var,
        ts_type=str(coltst),
        how=resample_how,
        min_num_obs=min_num_obs,
        inplace=True,
    )

    stat_data_ref.resample_time(
        var_name=var_ref,
        ts_type=str(coltst),
        how=resample_how,
        min_num_obs=min_num_obs,
        inplace=True,
    )

    # Save time indices of the observations and a mask of where it is NaN
    obs_idx = stat_data_ref[var_ref].index
    obs_isnan = stat_data_ref[var_ref].isnull()

    # now both StationData objects are in the same resolution, but they still
    # might have gaps in their time axis, thus concatenate them in a DataFrame,
    # which will merge the time index
    merged = pd.concat([stat_data_ref[var_ref], stat_data[var]], axis=1, keys=["ref", "data"])
    # Interpolate the model to the times of the observations
    # (for non-standard coltst it could be that 'resample_time'
    # has placed the model and observations at different time stamps)
    merged = merged.interpolate("index").reindex(obs_idx).loc[obs_idx]
    # Set to NaN at times when observations were NaN originally
    # (because the interpolation will interpolate the 'ref' column as well)
    merged.loc[obs_isnan] = np.nan
    # due to interpolation some model values may be NaN, where there is obs
    merged.loc[merged.data.isnull()] = np.nan
    # Ensure the whole timespan of the model is kept in "merged"
    stat_data[var].name = "tmp"
    merged = pd.concat([merged, stat_data[var]], axis=1)
    merged = merged[["ref", "data"]]

    grid_ts = merged["data"]
    obs_ts = merged["ref"]
    # invalidate model where obs is NaN (NB: maybe not needed any more?)
    obsnan = np.isnan(obs_ts.values)
    grid_ts[obsnan] = np.nan

    # now resample both to output frequency
    resampler = TimeResampler()
    obs_ts = resampler.resample(
        to_ts_type=ts_type,
        input_data=obs_ts,
        from_ts_type=coltst,
        how=resample_how,
        min_num_obs=min_num_obs,
    )

    grid_ts = resampler.resample(
        to_ts_type=ts_type,
        input_data=grid_ts,
        from_ts_type=coltst,
        how=resample_how,
        min_num_obs=min_num_obs,
    )
    # fill up missing time stamps
    return pd.concat([obs_ts, grid_ts], axis=1, keys=["ref", "data"])


def colocate_gridded_ungridded(
    data: GriddedData,
    data_ref: UngriddedDataContainer,
    ts_type=None,
    start=None,
    stop=None,
    filter_name=None,
    regrid_res_deg: float | RegridResDeg | None = None,
    regrid_scheme: str = "areaweighted",
    var_ref=None,
    update_baseyear_gridded=None,
    min_num_obs=None,
    colocate_time=False,
    use_climatology_ref=False,
    resample_how=None,
    **kwargs,
):
    """Colocate gridded with ungridded data (low level method)

    For high-level colocation see :class:`pyaerocom.colocation.Colocator`
    and :class:`pyaerocom.ColocationSetup`

    Note
    ----
    Uses the variable that is contained in input :class:`GriddedData` object
    (since these objects only contain a single variable). If this variable
    is not contained in observation data (or contained but using a different
    variable name) you may specify the obs variable to be used via input arg
    `var_ref`

    Parameters
    ----------
    data : GriddedData
        gridded data object (e.g. model results).
    data_ref : UngriddedDataContainer
        ungridded data object (e.g. observations).
    ts_type : str
        desired temporal resolution of colocated data (must be valid AeroCom
        ts_type str such as daily, monthly, yearly.).
    start : :obj:`str` or :obj:`datetime64` or similar, optional
        start time for colocation, if None, the start time of the input
        :class:`GriddedData` object is used.
    stop : :obj:`str` or :obj:`datetime64` or similar, optional
        stop time for colocation, if None, the stop time of the input
        :class:`GriddedData` object is used
    filter_name : str
        string specifying filter used (cf. :class:`pyaerocom.filter.Filter` for
        details). If None, then it is set to 'ALL-wMOUNTAINS', which
        corresponds to no filtering (world with mountains).
        Use ALL-noMOUNTAINS to exclude mountain sites.
    regrid_res_deg : int or dict, optional
        regrid resolution in degrees. If specified, the input gridded data
        object will be regridded in lon / lat dimension to the input
        resolution (if input is integer, both lat and lon are regridded to that
        resolution, if input is dict, use keys `lat_res_deg` and `lon_res_deg`
        to specify regrid resolutions, respectively).
    var_ref : :obj:`str`, optional
        variable against which data in arg `data` is supposed to be compared.
        If None, then the same variable is used (i.e. `data.var_name`).
    update_baseyear_gridded : int, optional
        optional input that can be set in order to re-define the time dimension
        in the gridded data object to be analysed. E.g., if the data object
        is a climatology (one year of data) that has set the base year of the
        time dimension to a value other than the specified input start / stop
        time this may be used to update the time in order to make colocation
        possible.
    min_num_obs : int or dict, optional
        minimum number of observations for resampling of time
    colocate_time : bool
        if True and if original time resolution of data is higher than desired
        time resolution (`ts_type`), then both datasets are colocated in time
        *before* resampling to lower resolution. NOTE: If there are missing data
        in the model, having this true will interpolate the missing model data!
    use_climatology_ref : ClimateConfig | bool, optional.
        Configuration for calculating the climatology. If set to a bool, this will not be done
    resample_how : str or dict
        string specifying how data should be aggregated when resampling in time.
        Default is "mean". Can also be a nested dictionary, e.g.
        resample_how={'daily': {'hourly' : 'max'}} would use the maximum value
        to aggregate from hourly to daily, rather than the mean.
    **kwargs
        additional keyword args (passed to
        :func:`UngriddedData.to_station_data_all`)

    Returns
    -------
    ColocatedData
        instance of colocated data

    Raises
    ------
    VarNotAvailableError
        if grid data variable is not available in ungridded data object
    AttributeError
        if instance of input :class:`UngriddedDataContainer` object contains more than
        one dataset
    TimeMatchError
        if gridded data time range does not overlap with input time range
    ColocationError
        if none of the data points in input :class:`UngriddedDataContainer` matches
        the input colocation constraints
    """
    if filter_name is None:
        filter_name = const.DEFAULT_REG_FILTER
    if data.proj_info is None:
        # reordering only implemented if no projection
        try:
            data.check_dimcoords_tseries()
        except DimensionOrderError:
            data.reorder_dimensions_tseries()

    var, var_aerocom = resolve_var_name(data)
    if var_ref is None:
        var_ref = var_aerocom
        var_ref_aerocom = var_aerocom
    else:
        var_ref_aerocom = const.VARS[var_ref].var_name_aerocom

    if var_ref not in data_ref.contains_vars:
        raise VarNotAvailableError(
            f"Variable {var_ref} is not available in ungridded "
            f"data (which contains {data_ref.contains_vars})"
        )
    elif len(data_ref.contains_datasets) > 1:
        raise AttributeError(
            f"Colocation can only be performed with ungridded data objects "
            f"that only contain a single dataset (input data contains: "
            f"{data_ref.contains_datasets}. Use method `extract_dataset` of "
            f"UngriddedData object to extract single datasets."
        )

    dataset_ref = data_ref.contains_datasets[0]

    if update_baseyear_gridded is not None:
        # update time dimension in gridded data
        data.base_year = update_baseyear_gridded

    # apply region filter to data
    regfilter = Filter(name=filter_name)
    data_ref = regfilter.apply(data_ref)
    data = regfilter.apply(data)

    # check time overlap and crop model data if needed
    start, stop = check_time_ival(data, start, stop)
    data = data.crop(time_range=(start, stop))

    if regrid_res_deg is not None:
        data = _regrid_gridded(data, regrid_scheme, regrid_res_deg)

    # Special ts_typs for which all stations with ts_type< are removed
    reduce_station_data_ts_type = ts_type

    ts_type_src_data = data.ts_type
    ts_type, ts_type_data = check_ts_type(data, ts_type)
    if not colocate_time and ts_type < ts_type_data:
        data = data.resample_time(str(ts_type), min_num_obs=min_num_obs, how=resample_how)
        ts_type_data = ts_type

    if isinstance(use_climatology_ref, ClimatologyConfig):  # pragma: no cover
        col_freq = use_climatology_ref.freq
        obs_start = use_climatology_ref.start
        obs_stop = use_climatology_ref.stop
    else:
        col_freq = str(ts_type)
        obs_start = start
        obs_stop = stop

    # colocation frequency
    col_tst = TsType(col_freq)

    latitude = data.latitude.points
    longitude = data.longitude.points
    if data.proj_info is None:
        lat_range = [np.min(latitude), np.max(latitude)]
        lon_range = [np.min(longitude), np.max(longitude)]
        # use only sites that are within model domain

        # filter_by_meta wipes is_vertical_profile
        data_ref = data_ref.filter_by_meta(latitude=lat_range, longitude=lon_range)
    else:
        # gridded data with projection,
        # add x/y information to ungridded
        for coord in data.cube.dim_coords:
            if coord.var_name == data.proj_info.x_axis:
                vals = coord.points
                xrange = (np.min(vals), np.max(vals))
            if coord.var_name == data.proj_info.y_axis:
                vals = coord.points
                yrange = (np.min(vals), np.max(vals))
        if xrange is None or yrange is None:
            raise VariableDefinitionError(
                f"x/y axis not found in cube: {data.proj_info.x_axis}, {data.proj_info.y_axis}"
            )
        data_ref = data_ref.filter_by_projection(data.proj_info.to_proj, xrange, yrange)

    # get timeseries from all stations in provided time resolution
    # (time resampling is done below in main loop)
    all_stats = data_ref.to_station_data_all(
        vars_to_convert=var_ref,
        start=obs_start,
        stop=obs_stop,
        by_station_name=True,
        ts_type_preferred=reduce_station_data_ts_type,
        **kwargs,
    )

    obs_stat_data = all_stats["stats"]
    ungridded_lons = all_stats["longitude"]
    ungridded_lats = all_stats["latitude"]

    if len(obs_stat_data) == 0:
        raise VarNotAvailableError(
            f"Variable {var_ref} is not available in specified time interval ({start}-{stop})"
        )

    # to read
    grid_stat_data = data.to_time_series(longitude=ungridded_lons, latitude=ungridded_lats)

    pd_freq = col_tst.to_pandas_freq()
    time_idx = make_datetime_index(start, stop, pd_freq)

    time_num = len(time_idx)
    stat_num = len(obs_stat_data)

    arr = np.full((2, time_num, stat_num), np.nan)

    lons = [np.nan] * stat_num
    lats = [np.nan] * stat_num
    alts = [np.nan] * stat_num
    station_names = [""] * stat_num
    station_types = [""] * stat_num
    station_display_names = [None] * stat_num

    data_ref_unit = None
    ts_type_src_ref = None
    data_unit = str(data.units)

    # loop over all stations and append to colocated data object
    for i, obs_stat in enumerate(obs_stat_data):
        # Add coordinates to arrays required for xarray.DataArray below
        lons[i] = obs_stat.longitude
        lats[i] = obs_stat.latitude
        alts[i] = obs_stat.altitude
        station_names[i] = obs_stat.station_name
        station_types[i] = getattr(obs_stat, "station_type", "")
        station_display_names[i] = getattr(obs_stat, "display_name", None)

        # ToDo: consider removing to keep ts_type_src_ref (this was probably
        # introduced for EBAS were the original data frequency is not constant
        # but can vary from site to site)
        if ts_type_src_ref is None:
            ts_type_src_ref = obs_stat["ts_type_src"]
        elif obs_stat["ts_type_src"] != ts_type_src_ref:
            spl = ts_type_src_ref.split(";")
            if obs_stat["ts_type_src"] not in spl:
                spl.append(obs_stat["ts_type_src"])
            ts_type_src_ref = ";".join(spl)

        if data_ref_unit is None:
            try:
                data_ref_unit = obs_stat["var_info"][var_ref]["units"]
            except KeyError as e:  # variable information or unit is not defined
                logger.exception(repr(e))
        try:
            unit = obs_stat["var_info"][var_ref]["units"]
        except Exception:
            unit = None
        if not unit == data_ref_unit:
            raise ValueError(
                f"Cannot perform colocation. "
                f"Ungridded data object contains different units ({var_ref})"
            )
        # get observations (Note: the index of the observation time series
        # is already in the specified frequency format, and thus, does not
        # need to be updated, for details (or if errors occur), cf.
        # UngriddedData.to_station_data, where the conversion happens)

        # get model station data
        grid_stat = grid_stat_data[i]

        try:
            if colocate_time:
                _df = _colocate_site_data_helper_timecol(
                    stat_data=grid_stat,
                    stat_data_ref=obs_stat,
                    var=var,
                    var_ref=var_ref,
                    ts_type=col_freq,
                    resample_how=resample_how,
                    min_num_obs=min_num_obs,
                    use_climatology_ref=use_climatology_ref,
                )
            else:
                _df = _colocate_site_data_helper(
                    stat_data=grid_stat,
                    stat_data_ref=obs_stat,
                    var=var,
                    var_ref=var_ref,
                    ts_type=col_freq,
                    resample_how=resample_how,
                    min_num_obs=min_num_obs,
                    use_climatology_ref=use_climatology_ref,
                )

            # this try/except block was introduced on 23/2/2021 as temporary fix from
            # v0.10.0 -> v0.10.1 as a result of multi-weekly obsdata (EBAS) that
            # can end up resulting in incorrect number of timestamps after resampling
            # (the error was discovered using EBASMC, concpm10, 2019 and colocation
            # frequency monthly)
            try:
                # assign the unified timeseries data to the colocated data array
                arr[0, :, i] = _df["ref"].values
                arr[1, :, i] = _df["data"].values
            except ValueError:
                try:
                    mask = _df.index.intersection(time_idx)
                    _df = _df.loc[mask]
                    arr[0, :, i] = _df["ref"].values
                    arr[1, :, i] = _df["data"].values
                except ValueError as e:
                    logger.warning(
                        f"Failed to colocate time for station {obs_stat.station_name}. "
                        f"This station will be skipped (error: {e})"
                    )
        except TemporalResolutionError as e:
            # resolution of obsdata is too low
            logger.warning(
                f"{var_ref} data from site {obs_stat.station_name} will "
                f"not be added to ColocatedData. Reason: {e}"
            )
    try:
        revision = data_ref.get_data_revision(dataset_ref)
    except MetaDataError:
        revision = "MULTIPLE"
    except Exception:
        revision = "n/a"

    files = [os.path.basename(x) for x in data.from_files]

    meta = {
        "data_source": [dataset_ref, data.data_id],
        "var_name": [var_ref_aerocom, var_aerocom],
        "var_name_input": [var_ref, var],
        "ts_type": col_freq,  # will be updated below if resampling
        "filter_name": filter_name,
        "ts_type_src": [ts_type_src_ref, ts_type_src_data],
        "var_units": [data_ref_unit, data_unit],
        "data_level": 3,
        "revision_ref": revision,
        "from_files": files,
        "from_files_ref": None,
        "colocate_time": colocate_time,
        "obs_is_clim": (True if isinstance(use_climatology_ref, ClimatologyConfig) else False),
        "pyaerocom": pya_ver,
        "min_num_obs": min_num_obs,
        "resample_how": resample_how,
    }

    # create coordinates of DataArray
    coords = {
        "data_source": meta["data_source"],
        "time": time_idx,
        "station_name": station_names,
        "station_type": ("station_name", station_types),
        "latitude": ("station_name", lats),
        "longitude": ("station_name", lons),
        "altitude": ("station_name", alts),
    }
    if any(x is not None for x in station_display_names):
        coords["station_display_name"] = ("station_name", station_display_names)

    dims = ["data_source", "time", "station_name"]
    coldata = ColocatedData(data=arr, coords=coords, dims=dims, name=var, attrs=meta)

    # add correct units for lat / lon dimensions
    coldata.latitude.attrs["standard_name"] = data.latitude.standard_name
    coldata.latitude.attrs["units"] = str(data.latitude.units)

    coldata.longitude.attrs["standard_name"] = data.longitude.standard_name
    coldata.longitude.attrs["units"] = str(data.longitude.units)

    return coldata


def correct_model_stp_coldata(coldata, p0=None, t0=273.15, inplace=False):
    """Correct modeldata in colocated data object to STP conditions

    Note
    ----
    BETA version, quite unelegant coded (at 8pm 3 weeks before IPCC deadline),
    but should do the job for 2010 monthly colocated data files (AND NOTHING
    ELSE)!

    """
    if coldata.ndim != 3:
        raise NotImplementedError("Can only handle 3D coldata so far...")
    elif not coldata.ts_type == "monthly" or not len(coldata.time) == 12:
        raise NotImplementedError(
            "Can only handle monthly colocated data files "
            "so far (since ERA5 temps are only available) "
            "in monthly resolution"
        )
    startyr = pd.Timestamp(coldata.start).year
    stopyr = pd.Timestamp(coldata.stop).year
    if not all([x == 2010 for x in (startyr, stopyr)]):
        raise NotImplementedError("Can only handle 2010 monthly data so far")

    if not inplace:
        coldata = coldata.copy()

    arr = coldata.data

    coords = zip(
        arr.latitude.values,
        arr.longitude.values,
        arr.altitude.values,
        arr.station_name.values,
    )
    if p0 is None:
        p0 = pressure()  # STD conditions sea level
    logger.info("Correcting model data in ColocatedData instance to STP")
    cfacs = []
    meantemps = []
    mintemps = []
    maxtemps = []
    ps = []
    with xr.open_dataset(const.ERA5_SURFTEMP_FILE, decode_timedelta=True)["t2m"] as temp:
        for i, (lat, lon, alt, name) in enumerate(coords):
            logger.info(name, ", Lat", lat, ", Lon", lon)
            p = pressure(alt)
            logger.info("Alt", alt)
            logger.info("P=", p / 100, "hPa")

            ps.append(p / 100)

            temps = temp.sel(latitude=lat, longitude=lon, method="nearest").data

            meantemps.append(temps.mean())
            mintemps.append(temps.min())
            maxtemps.append(temps.min())

            if not len(temps) == len(arr.time):
                raise NotImplementedError("Check timestamps")
            logger.info("Mean Temp: ", temps.mean() - t0, " C")

            corrfacs = (p0 / p) * (temps / t0)

            logger.info("Corr fac:", corrfacs.mean(), "+/-", corrfacs.std())

            cfacs.append(corrfacs.mean())

            # mularr = xr.DataArray(corrfacs)

            if not arr.station_name.values[i] == name:
                raise Exception
            elif not arr.dims[1] == "time":
                raise Exception
            arr[1, :, i] *= corrfacs

    cfacs = np.asarray(cfacs)

    logger.info("Min: ", cfacs.min())
    logger.info("Mean: ", cfacs.mean())
    logger.info("Max: ", cfacs.max())
    coldata.data.attrs["Model_STP_corr"] = True

    newcoords = dict(
        pres=("station_name", ps),
        temp_mean=("station_name", meantemps),
        temp_min=("station_name", mintemps),
        temp_max=("station_name", maxtemps),
        stp_corrfac_mean=("station_name", cfacs),
    )

    coldata.data = coldata.data.assign_coords(newcoords)

    info_str = (
        "Correction factors to convert model data from ambient to "
        "STP were computed using corrfac=(p0/p)*(T/T0) with T0=273K "
        "and p0=1013 hPa and p is the pressure at the station location "
        "(which was computed assuming a standard atmosphere and using "
        "the station altitude) and T is the 2m surface temperature at "
        "the station, applied on a monthly basis and estimated using "
        "ERA5 data"
    )

    coldata.data["pres"].attrs["units"] = "hPa"
    coldata.data["temp_mean"].attrs["units"] = "K"
    coldata.data["temp_min"].attrs["units"] = "K"
    coldata.data["temp_max"].attrs["units"] = "K"

    coldata.data.attrs["Model_STP_corr"] = True
    coldata.data.attrs["Model_STP_corr_info"] = info_str
    return coldata



from __future__ import annotations

import pytest

from pyaerocom.units import UnitConversionError
from pyaerocom.units.units_helpers import (
    convert_unit,
    get_unit_conversion_fac,
)


@pytest.mark.parametrize(
    "from_unit,to_unit,var_name,result",
    [
        ("ug m-3", "ug/m3", None, 1),
        ("ug m-3", "ug/m3", "concso2", 1),
        ("mg m-3", "ug m-3", "concso2", 1e3),
        ("ug S/m3", "mg m-3", "concso2", 1.9979e-3),
    ],
)
def test_convert_unit(from_unit: str, to_unit: str, var_name: str, result: float):
    converted = convert_unit(1, from_unit, to_unit, var_name)
    assert converted == pytest.approx(result, rel=1e-2)


@pytest.mark.parametrize(
    "from_unit,to_unit,var_name,ts_type,result",
    [
        ("kg (2m)-2", "kg/m2", None, None, 0.25),
        ("mm", "mm d-1", "prmm", "daily", 1),
        ("mm", "mm d-1", "prmm", "hourly", 24),
        ("mg m-2", "ug m-2 d-1", "wetoxs", "hourly", 24e3),
        ("mg m-2", "mg m-2 d-1", "wetoxs", "hourly", 24),
        ("mg m-2", "mg m-2/d", "wetoxs", "daily", 1),
        ("mg m-2", "mg m-2 d-1", "wetoxs", "daily", 1),
        ("mg m-2", "ug mm-2", None, None, 1e-3),
        ("mg", "ug", None, None, 1000),
        ("1", "1", None, None, 1),
    ],
)
def test_get_unit_conversion_fac(
    from_unit: str, to_unit: str, var_name: str | None, ts_type: str | None, result: float
):
    conversion_fac = get_unit_conversion_fac(from_unit, to_unit, var_name, ts_type)
    assert conversion_fac == pytest.approx(result, rel=1e-3)


@pytest.mark.parametrize(
    "from_unit,to_unit,var_name",
    [
        ("kg m-2", "mm", "pr"),
        ("kg m-2", "mm", "od550aer"),
        ("1", "ug", None),
    ],
)
def test_get_unit_conversion_fac_error(from_unit: str, to_unit: str, var_name: str | None):
    with pytest.raises(UnitConversionError) as e:
        get_unit_conversion_fac(from_unit, to_unit, var_name)
    assert str(e.value) == f"failed to convert unit from {from_unit} to {to_unit}"



from pyaerocom.units.logging import LoggingCallback
from pyaerocom.units import Unit
import logging


def test_callback_1(caplog):
    caplog.set_level(logging.INFO)

    Unit("1").convert(1, "1", callback=LoggingCallback())

    assert "Successfully converted unit of variable" not in caplog.text


def test_callback_2(caplog):
    caplog.set_level(logging.DEBUG)

    Unit("1").convert(1, "1", callback=LoggingCallback())

    assert "Successfully converted unit of variable" in caplog.text


def test_callback_3(caplog):
    caplog.set_level(logging.INFO)

    Unit("1").convert(1, "2", callback=LoggingCallback())

    assert "Successfully converted unit of variable" in caplog.text



import pytest
import numpy as np
from datetime import timedelta
from pyaerocom.units.datetime import seconds_in_periods


@pytest.mark.parametrize(
    "date,ts_type,days",
    [
        pytest.param("2000-02-18", "yearly", 366, id="leap year"),
        pytest.param("2001-02-18", "yearly", 365, id="lon leap year"),
        pytest.param("2000-02-18", "monthly", 29, id="February leap year"),
        pytest.param("2001-02-18", "monthly", 28, id="February non leap year"),
        pytest.param("2001-02-18", "daily", 1, id="one day"),
    ],
)
def test_seconds_in_periods(date, ts_type, days):
    seconds = timedelta(days=days) / timedelta(seconds=1)
    assert seconds_in_periods(np.datetime64(date), ts_type) == seconds
